# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    blockchain_info client
#    © 2017-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "blockchaininfo"

_logger = logging.getLogger(__name__)


class BlockchainInfoClient(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(self, cmd, parameter="", variables=None, method="get"):
        url_path = cmd
        if parameter:
            url_path += "/" + parameter
        return self.request(url_path, variables, method=method)

    def getbalance(self, addresslist):
        addresses = {"active": "|".join(addresslist)}
        res = self.compose_request("balance", variables=addresses)
        balance = 0
        for address in res:
            balance += res[address]["final_balance"]
        return balance

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        utxos = []
        variables = {"active": address, "limit": 1000}
        res = self.compose_request("unspent", variables=variables)
        if len(res["unspent_outputs"]) > 299:
            _logger.info(
                "BlockchainInfoClient: Large number of outputs for address %s, "
                "UTXO list may be incomplete" % address
            )
        res["unspent_outputs"].sort(key=lambda x: x["confirmations"])
        for utxo in res["unspent_outputs"]:
            if utxo["tx_hash_big_endian"] == after_txid:
                break
            utxos.append(
                {
                    "address": address,
                    "txid": utxo["tx_hash_big_endian"],
                    "confirmations": utxo["confirmations"],
                    "output_n": utxo["tx_output_n"],
                    "input_n": utxo["tx_index"],
                    "block_height": None,
                    "fee": None,
                    "size": 0,
                    "value": int(round(utxo["value"] * self.units, 0)),
                    "script": utxo["script"],
                    "date": None,
                }
            )
        return utxos[::-1][:limit]

    def gettransaction(self, txid, latest_block=None):
        self.latest_block = self.latest_block if not latest_block else latest_block
        tx = self.compose_request("rawtx", txid)
        rawtx = self.getrawtransaction(txid)
        t = BitcoinTransaction.parse_hex(
            rawtx, strict=self.strict, network=self.network
        )
        input_total = 0
        for n, i in enumerate(t.inputs):
            if "prev_out" in tx["inputs"][n]:
                i.value = (
                    0
                    if not tx["inputs"][n]["prev_out"]
                    else tx["inputs"][n]["prev_out"]["value"]
                )
                input_total += i.value
        for n, o in enumerate(t.outputs):
            o.spent = tx["out"][n]["spent"]
        if "block_height" in tx and tx["block_height"]:
            if not self.latest_block:
                self.latest_block = self.blockcount()
            t.status = "confirmed"
            t.date = datetime.utcfromtimestamp(tx["time"])
            t.block_height = tx["block_height"]
            t.confirmations = 1
            if self.latest_block > t.block_height:
                t.confirmations = self.latest_block - t.block_height
        else:
            t.status = "unconfirmed"
            t.confirmations = 0
            t.date = None
        t.rawtx = bytes.fromhex(rawtx)
        t.size = tx["size"]
        t.network_name = self.network
        t.locktime = tx["lock_time"]
        t.version_int = tx["ver"]
        t.version = tx["ver"].to_bytes(4, "big")
        t.input_total = input_total
        t.fee = 0
        if t.input_total:
            t.fee = t.input_total - t.output_total
        return t

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txs = []
        txids = []
        variables = {"limit": 100}
        res = self.compose_request("rawaddr", address, variables=variables)
        self.latest_block = (
            self.blockcount() if not self.latest_block else self.latest_block
        )
        for tx in res["txs"]:
            if tx["hash"] not in txids:
                txids.insert(0, tx["hash"])
        if after_txid:
            txids = txids[txids.index(after_txid) + 1 :]
        for txid in txids[:limit]:
            t = self.gettransaction(txid, latest_block=self.latest_block)
            t.confirmations = (
                0 if not t.block_height else self.latest_block - t.block_height
            )
            txs.append(t)
        return txs

    def getrawtransaction(self, txid):
        return self.compose_request("rawtx", txid, {"format": "hex"})

    # def sendrawtransaction()

    # def estimatefee()

    def blockcount(self):
        return self.compose_request("latestblock")["height"]

    def mempool(self, txid=""):
        if txid:
            tx = self.compose_request("rawtx", txid)
            if "block_height" not in tx:
                return [tx["hash"]]
        else:
            txs = self.compose_request(
                "unconfirmed-transactions", variables={"format": "json"}
            )
            return [tx["hash"] for tx in txs["txs"]]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        bd = self.compose_request("rawblock", str(blockid))
        if parse_transactions:
            txs = []
            self.latest_block = (
                self.blockcount() if not self.latest_block else self.latest_block
            )
            for tx in bd["tx"][(page - 1) * limit : page * limit]:
                # try:
                txs.append(
                    self.gettransaction(tx["hash"], latest_block=self.latest_block)
                )
                # except Exception as e:
                #     _logger.error("Could not parse tx %s with error %s" % (tx['hash'], e))
        else:
            txs = [tx["hash"] for tx in bd["tx"]]

        block = {
            "bits": bd["bits"],
            "depth": None,
            "block_hash": bd["hash"],
            "height": bd["height"],
            "merkle_root": bd["mrkl_root"],
            "nonce": abs(bd["nonce"]),
            "prev_block": bd["prev_block"],
            "time": bd["time"],
            "tx_count": len(bd["tx"]),
            "txs": txs,
            "version": bd["ver"],
            "page": page,
            "pages": None
            if not limit
            else int(len(bd["tx"]) // limit) + (len(bd["tx"]) % limit > 0),
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        return self.compose_request("rawblock", str(blockid), {"format": "hex"})

    # def isspent(self, txid, index):

    def getinfo(self):
        import json

        import requests

        info = json.loads(
            requests.get("https://api.blockchain.info/stats", timeout=self.timeout).text
        )
        unconfirmed = self.compose_request("q", "unconfirmedcount")
        return {
            "blockcount": info["n_blocks_total"],
            "chain": "",
            "difficulty": info["difficulty"],
            "hashrate": int(float(info["hash_rate"] * 10**9)),
            "mempool_size": unconfirmed,
        }



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#
#    Create a multisig 2-of-3 wallet with Mnemonic passphrase keys, so wallet contains 3 keys and 2 signatures are
#    needed to sign a transaction / send a payment.
#
#    Transaction are created and signed with 1 signature on the online PC, on the other offline PC the transaction is
#    signed with a second private key. The third key is a stored on a paper in case one of the others keys is lost.
#
#    © 2017 - 2019 December - 1200 Web Development <http://1200wd.com/>
#

from __future__ import print_function

from pprint import pprint

from fluxwallet.keys import HDKey
from fluxwallet.mnemonic import Mnemonic
from fluxwallet.wallet import Wallet, wallet_exists

WALLET_NAME = "Multisig-2of3"
NETWORK = "testnet"
KEY_STRENGTH = 128  # Remove this line to use the default 256 bit key strength
SIGNATURES_REQUIRED = 2
WITNESS_TYPE = "segwit"  # Witness type can be legacy, p2sh-segwit or segwit

# from fluxwallet.wallets import wallet_delete_if_exists
# wallet_delete_if_exists(WALLET_NAME, force=True)

if not wallet_exists(WALLET_NAME):
    # Define cosigners, format (name, key_type, [password], wallet)
    cosigners = [
        ("Offline PC", "bip32", "password"),
        ("Online PC", "bip32", ""),
        ("Paper backup", "single", ""),
    ]

    print(
        "We will generate 3 private keys, to sign and send a transaction 2 keys are needed:"
        "\n- With 1 private key a wallet on this Offline PC is created"
        "\n- Use private key 2 to create a wallet on the Online PC"
        "\n- Store key 3 on a Paper in a safe in case one of the PC's is not available anymore"
    )
    key_lists = {}
    w_id = 0
    for cosigner in cosigners:
        print("\n")
        words = Mnemonic().generate(KEY_STRENGTH)
        password = ""
        if cosigner[2] == "password":
            password = input("Please give password for cosigner '%s': " % cosigner[0])
        seed = Mnemonic().to_seed(words, password)
        hdkey = HDKey.from_seed(
            seed, network=NETWORK, key_type=cosigner[1], witness_type=WITNESS_TYPE
        )
        if cosigner[1] == "bip32":
            public_account = hdkey.public_master_multisig(witness_type=WITNESS_TYPE)
        else:
            public_account = hdkey
        print(
            "Key for cosigner '%s' generated. Please store both passphrase and password carefully!"
            % cosigner[0]
        )
        print("Passphrase: %s" % words)
        print("Password: %s" % ("*" * len(password)))
        print("Share this public key below with other cosigner")
        print("Public key: %s" % public_account.wif_public())

        for w in cosigners:
            if cosigner[0] == w[0]:
                addkey = hdkey
            else:
                addkey = public_account.public()
            if w[0] not in key_lists:
                key_lists[w[0]] = []
            if addkey not in key_lists[w[0]]:
                key_lists[w[0]].append(addkey)

    offline_wallet = Wallet.create(
        WALLET_NAME,
        key_lists["Offline PC"],
        sigs_required=SIGNATURES_REQUIRED,
        witness_type=WITNESS_TYPE,
        network=NETWORK,
    )
    offline_wallet.new_key()

    print("\n\nA multisig wallet has been created on this system")
    offline_wallet.info()

    print("\n---> Please create a wallet on your Online PC like this:")
    print("from fluxwallet.wallets import Wallet")
    print("from fluxwallet.keys import HDKey")
    print("")
    print("key_list = [")
    for key in key_lists["Online PC"]:
        if key.key_type == "single":
            print(
                "     HDKey('%s', key_type='single', witness_type='%s')"
                % (key.wif_private(), WITNESS_TYPE)
            )
        else:
            print("     '%s'," % key.wif_private())
    print("]")
    print(
        "wlt = Wallet.create('%s', key_list, sigs_required=2, witness_type='%s', network='%s')"
        % (WALLET_NAME, WITNESS_TYPE, NETWORK)
    )
    print("wlt.get_key()")
    print("wlt.info()")
else:
    from fluxwallet.config.config import FLUXWALLET_VERSION, FW_DATABASE_DIR

    online_wallet = Wallet(
        WALLET_NAME, db_uri=FW_DATABASE_DIR + "/fluxwallet.tmp.sqlite"
    )
    online_wallet.utxos_update()
    online_wallet.info()
    utxos = online_wallet.utxos()
    if utxos:
        print("\nNew unspent outputs found!")
        print(
            "Now a new transaction will be created to sweep this wallet and send bitcoins to a testnet faucet"
        )
        send_to_address = "n2eMqTT929pb1RDNuqEnxdaLau1rxy3efi"
        t = online_wallet.sweep(send_to_address, min_confirms=0)
        print(t.raw_hex())
        print(
            "Now copy-and-paste the raw transaction hex to your Offline PC and sign it there with a second signature:"
        )
        print("\nfrom fluxwallet.wallets import Wallet")
        print("")
        print("wlt = Wallet('%s')" % WALLET_NAME)
        print("utxos = ", end="")
        pprint(utxos)
        print("")
        print("wlt.utxos_update(utxos=utxos)")
        print("t = wlt.transaction_import_raw('%s')" % t.raw_hex())
        print("t.sign()")
        print("")
        print(
            "# Push the following raw transaction to the blockchain network on any online PC:"
        )
        print("print(t.raw_hex())")
    else:
        print(
            "\nPlease send funds to %s, so we can create a transaction"
            % online_wallet.get_key().address
        )
        print("\nRestart this program when funds are send...")



import logging

from sqlalchemy import or_, select
from sqlalchemy.ext.asyncio import AsyncSession

from fluxwallet.config.config import DEFAULT_WITNESS_TYPE
from fluxwallet.db_new import Db, DbKey, DbNetwork
from fluxwallet.encoding import *
from fluxwallet.keys import Address, HDKey
from fluxwallet.networks import Network
from fluxwallet.values import Value
from fluxwallet.wallet.errors import WalletError

DEFAULT_WITNESS_TYPE: str

_logger = logging.getLogger(__name__)


class WalletKey:
    """
    Used as attribute of Wallet class. Contains HDKey class, and adds extra wallet related information such as
    key ID, name, path and balance.

    All WalletKeys are stored in a database
    """

    @staticmethod
    async def from_db_key(
        key_id: int, async_session: AsyncSession, *, hdkey: HDKey | None = None
    ):
        async with async_session as session:
            if db_key := await session.get(DbKey, key_id):
                await db_key.awaitable_attrs.wallet
                return WalletKey(db_key, hdkey)
            else:
                raise WalletError(f"Key with id {key_id} not found")

    @staticmethod
    async def from_hdkey(
        name: str,
        wallet_id: int,
        async_session: AsyncSession,
        *,
        key: HDKey | Address | str | bytes | None = None,
        account_id: int = 0,
        network: str | None = None,
        change: int = 0,
        purpose: int = 44,
        parent_id: int = 0,
        path: str = "m",
        key_type: str | None = None,
        encoding: str | None = None,
        witness_type: str | None = DEFAULT_WITNESS_TYPE,
        multisig: bool = False,
        cosigner_id: int | None = None,
    ):
        """
        Create WalletKey from a HDKey object or key.

        Normally you don't need to call this method directly. Key creation is handled by the Wallet class.

        >>> w = wallet_create_or_open('hdwalletkey_test')
        >>> wif = 'xprv9s21ZrQH143K2mcs9jcK4EjALbu2z1N9qsMTUG1frmnXM3NNCSGR57yLhwTccfNCwdSQEDftgjCGm96P29wGGcbBsPqZH85iqpoHA7LrqVy'
        >>> db_key = WalletKey.from_key('import_key', w.wallet_id, w._session, wif)
        >>> self.dbkey.address
        '1MwVEhGq6gg1eeSrEdZom5bHyPqXtJSnPg'
        >>> db_key # doctest:+ELLIPSIS
        <WalletKey(key_id=..., name=import_key, wif=xprv9s21ZrQH143K2mcs9jcK4EjALbu2z1N9qsMTUG1frmnXM3NNCSGR57yLhwTccfNCwdSQEDftgjCGm96P29wGGcbBsPqZH85iqpoHA7LrqVy, path=m)>

        :param name: New key name
        :type name: str
        :param wallet_id: ID of wallet where to store key
        :type wallet_id: int
        :param session: Required Sqlalchemy Session object
        :type session: sqlalchemy.orm.session.Session
        :param key: Optional key in any format accepted by the HDKey class
        :type key: str, int, byte, HDKey
        :param account_id: Account ID for specified key, default is 0
        :type account_id: int
        :param network: Network of specified key
        :type network: str
        :param change: Use 0 for normal key, and 1 for change key (for returned payments)
        :type change: int
        :param purpose: BIP0044 purpose field, default is 44
        :type purpose: int
        :param parent_id: Key ID of parent, default is 0 (no parent)
        :type parent_id: int
        :param path: BIP0044 path of given key, default is 'm' (masterkey)
        :type path: str
        :param key_type: Type of key, single or BIP44 type
        :type key_type: str
        :param encoding: Encoding used for address, i.e.: base58 or bech32. Default is base58
        :type encoding: str
        :param witness_type: Witness type used when creating transaction script: legacy, p2sh-segwit or segwit.
        :type witness_type: str
        :param multisig: Specify if key is part of multisig wallet, used for create keys and key representations such as WIF and addreses
        :type multisig: bool
        :param cosigner_id: Set this if you would like to create keys for other cosigners.
        :type cosigner_id: int

        :return WalletKey: WalletKey object
        """
        key_is_address = False
        if isinstance(key, HDKey):
            hd_key = key
            if network is None:
                network = hd_key.network.name
            elif network != hd_key.network.name:
                raise WalletError(
                    "Specified network and key network should be the same"
                )
        elif isinstance(key, Address):
            hd_key = key
            key_is_address = True
            if network is None:
                network = hd_key.network.name
            elif network != hd_key.network.name:
                raise WalletError(
                    "Specified network and key network should be the same"
                )
        else:  # str | bytes
            if network is None:
                network = DEFAULT_NETWORK
            hd_key = HDKey(import_key=key, network=network)

        if not encoding and witness_type:
            encoding = get_encoding_from_witness(witness_type)

        script_type = script_type_default(witness_type, multisig)

        async with async_session as session:
            if not key_is_address:
                res = await session.scalars(
                    select(DbKey).filter(
                        DbKey.wallet_id == wallet_id,
                        DbKey.wif
                        == hd_key.wif(
                            witness_type=witness_type,
                            multisig=multisig,
                            is_private=True,
                        ),
                    )
                )

                existing_db_key = res.first()

                if existing_db_key:
                    _logger.warning(
                        "Key already exists in this wallet. Key ID: %d"
                        % existing_db_key.id
                    )
                    await existing_db_key.awaitable_attrs.wallet
                    return WalletKey(existing_db_key, hd_key)

                if key_type != "single" and hd_key.depth != len(path.split("/")) - 1:
                    if path == "m" and hd_key.depth > 1:
                        path = "M"

                address = hd_key.address(encoding=encoding, script_type=script_type)

                res = await session.scalars(
                    select(DbKey).where(
                        DbKey.wallet_id == wallet_id,
                        or_(
                            DbKey.public == hd_key.public_byte,
                            DbKey.wif
                            == hd_key.wif(
                                witness_type=witness_type,
                                multisig=multisig,
                                is_private=False,
                            ),
                            DbKey.address == address,
                        ),
                    )
                )
                db_key = res.first()

                if db_key:
                    db_key.wif = hd_key.wif(
                        witness_type=witness_type, multisig=multisig, is_private=True
                    )
                    db_key.is_private = True
                    db_key.private = hd_key.private_byte
                    db_key.public = hd_key.public_byte
                    db_key.path = path

                    await session.commit()
                    await db_key.awaitable_attrs.wallet

                    return WalletKey(db_key, hd_key)

                new_db_key = DbKey(
                    name=name[:80],
                    wallet_id=wallet_id,
                    public=hd_key.public_byte,
                    private=hd_key.private_byte,
                    purpose=purpose,
                    account_id=account_id,
                    depth=hd_key.depth,
                    change=change,
                    address_index=hd_key.child_index,
                    wif=hd_key.wif(
                        witness_type=witness_type, multisig=multisig, is_private=True
                    ),
                    address=address,
                    parent_id=parent_id,
                    compressed=hd_key.compressed,
                    is_private=hd_key.is_private,
                    path=path,
                    key_type=key_type,
                    network_name=network,
                    encoding=encoding,
                    cosigner_id=cosigner_id,
                )
            else:
                res = await session.scalars(
                    select(DbKey).where(
                        DbKey.wallet_id == wallet_id, DbKey.address == hd_key.address
                    )
                )
                existing_db_key = res.first()

                if existing_db_key:
                    _logger.warning(
                        "Key with ID %s already exists" % existing_db_key.id
                    )

                    await existing_db_key.awaitable_attrs.wallet
                    return WalletKey(existing_db_key, hd_key)

                new_db_key = DbKey(
                    name=name[:80],
                    wallet_id=wallet_id,
                    purpose=purpose,
                    account_id=account_id,
                    depth=hd_key.depth,
                    change=change,
                    address=hd_key.address(),
                    parent_id=parent_id,
                    compressed=hd_key.compressed,
                    is_private=False,
                    path=path,
                    key_type=key_type,
                    network_name=network,
                    encoding=encoding,
                    cosigner_id=cosigner_id,
                )

            await session.merge(DbNetwork(name=network))
            session.add(new_db_key)

            await session.commit()
            await new_db_key.awaitable_attrs.wallet

        return WalletKey(new_db_key, hd_key)

    # def _commit(self):
    #     try:
    #         self._session.commit()
    #     except Exception:
    #         self._session.rollback()
    #         raise

    def __init__(self, db_key: DbKey, hdkey: HDKey | None = None):
        """
        Initialize WalletKey with specified ID, get information from database.

        :param key_id: ID of key as mentioned in database
        :type key_id: int
        :param session: Required Sqlalchemy Session object
        :type session: sqlalchemy.orm.session.Session
        :param hdkey: Optional HDKey object. Specify HDKey object if available for performance
        :type hdkey: HDKey

        """
        self.dbkey = db_key
        self.hdkey = hdkey

        # fix this
        if hdkey and isinstance(hdkey, HDKey):
            assert not self.dbkey.public or self.dbkey.public == hdkey.public_byte
            assert not self.dbkey.private or self.dbkey.private == hdkey.private_byte

        self.key_id = self.dbkey.id
        self._name = self.dbkey.name
        self.wallet_id = self.dbkey.wallet_id
        self.key_public = None if not self.dbkey.public else self.dbkey.public
        self.key_private = None if not self.dbkey.private else self.dbkey.private
        self.account_id = self.dbkey.account_id
        self.change = self.dbkey.change
        self.address_index = self.dbkey.address_index
        self.wif = self.dbkey.wif
        self.address = self.dbkey.address
        self._balance = self.dbkey.balance
        self.purpose = self.dbkey.purpose
        self.parent_id = self.dbkey.parent_id
        self.is_private = self.dbkey.is_private
        self.path = self.dbkey.path
        self.wallet = self.dbkey.wallet
        self.network_name = self.dbkey.network_name

        if not self.network_name:
            self.network_name = self.dbkey.wallet.network_name

        self.network = Network(self.network_name)
        self.depth = self.dbkey.depth
        self.key_type = self.dbkey.key_type
        self.compressed = self.dbkey.compressed
        self.encoding = self.dbkey.encoding
        self.cosigner_id = self.dbkey.cosigner_id
        self.used = self.dbkey.used

    def __repr__(self):
        return "<WalletKey(key_id=%d, name=%s, wif=%s, path=%s)>" % (
            self.key_id,
            self.name,
            self.wif,
            self.path,
        )

    @property
    def name(self):
        """
        Return name of wallet key

        :return str:
        """
        return self._name

    # @name.setter
    # def name(self, value):
    #     """
    #     Set key name, update in database

    #     :param value: Name for this key
    #     :type value: str

    #     :return str:
    #     """

    #     self._name = value
    #     self.dbkey.name = value
    #     self._commit()

    def key(self) -> HDKey:
        """
        Get HDKey object for current WalletKey

        :return HDKey:
        """

        self.hdkey = None
        if self.key_type == "multisig":
            self.hdkey = []
            for kc in self.dbkey.multisig_children:
                self.hdkey.append(
                    HDKey.from_wif(
                        kc.child_key.wif,
                        network=kc.child_key.network_name,
                        compressed=self.compressed,
                    )
                )
        if self.hdkey is None and self.wif:
            self.hdkey = HDKey.from_wif(
                self.wif, network=self.network_name, compressed=self.compressed
            )
        return self.hdkey

    def balance(self, as_string=False):
        """
        Get total value of unspent outputs

        :param as_string: Specify 'string' to return a string in currency format
        :type as_string: bool

        :return float, str: Key balance
        """

        if as_string:
            return Value.from_satoshi(self._balance, network=self.network).str_unit()
        else:
            return self._balance

    def public(self):
        """
        Return current key as public WalletKey object with all private information removed

        :return WalletKey:
        """
        pub_key = self
        pub_key.is_private = False
        pub_key.key_private = None

        if self.key():
            pub_key.wif = self.key().wif()
        if self.hdkey:
            self.hdkey = pub_key.hdkey.public()

        self.dbkey = None

        return pub_key

    def as_dict(self, include_private=False):
        """
        Return current key information as dictionary

        :param include_private: Include private key information in dictionary
        :type include_private: bool

        """

        kdict = {
            "id": self.key_id,
            "key_type": self.key_type,
            "network": self.network.name,
            "is_private": self.is_private,
            "name": self.name,
            "key_public": "" if not self.key_public else self.key_public.hex(),
            "account_id": self.account_id,
            "parent_id": self.parent_id,
            "depth": self.depth,
            "change": self.change,
            "address_index": self.address_index,
            "address": self.address,
            "encoding": self.encoding,
            "path": self.path,
            "balance": self.balance(),
            "balance_str": self.balance(as_string=True),
        }
        if include_private:
            kdict.update(
                {
                    "key_private": self.key_private.hex(),
                    "wif": self.wif,
                }
            )
        return kdict



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    litecoinblockexplorer.net Client
#    © 2019-2022 - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "litecoinblockexplorer"
REQUEST_LIMIT = 50

_logger = logging.getLogger(__name__)


class LitecoinBlockexplorerClient(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self, category, data, cmd="", variables=None, method="get", offset=0
    ):
        url_path = category
        if data:
            url_path += "/" + data + ("" if not cmd else "/" + cmd)
        if variables is None:
            variables = {}
        variables.update({"from": offset, "to": offset + REQUEST_LIMIT})
        return self.request(url_path, variables, method=method)

    def _convert_to_transaction(self, tx):
        if tx["confirmations"]:
            status = "confirmed"
        else:
            status = "unconfirmed"
        fees = (
            None if "fees" not in tx else int(round(float(tx["fees"]) * self.units, 0))
        )
        value_in = (
            0
            if "valueIn" not in tx
            else int(round(float(tx["valueIn"]) * self.units, 0))
        )
        isCoinbase = False
        if "isCoinBase" in tx and tx["isCoinBase"]:
            isCoinbase = True
        txdate = None
        if "blocktime" in tx:
            txdate = datetime.utcfromtimestamp(tx["blocktime"])
        t = BitcoinTransaction(
            locktime=tx.get("locktime", 0),
            version=tx["version"],
            network=self.network,
            fee=fees,
            txid=tx["txid"],
            date=txdate,
            confirmations=tx["confirmations"],
            block_height=tx["blockheight"],
            status=status,
            input_total=int(round(float(value_in) * self.units, 0)),
            coinbase=isCoinbase,
            output_total=int(round(float(tx["valueOut"]) * self.units, 0)),
            size=len(tx["hex"]) // 2,
        )
        for ti in tx["vin"]:
            if isCoinbase:
                t.add_input(
                    prev_txid=32 * b"\0",
                    output_n=4 * b"\xff",
                    unlocking_script=ti["coinbase"],
                    index_n=ti["n"],
                    script_type="coinbase",
                    sequence=ti.get("sequence", 0),
                    value=0,
                )
            else:
                value = int(round(float(ti["value"] or 0) * self.units, 0))
                us = "" if "hex" not in ti["scriptSig"] else ti["scriptSig"]["hex"]
                t.add_input(
                    prev_txid=ti["txid"],
                    output_n=ti["vout"],
                    unlocking_script=us,
                    index_n=ti["n"],
                    value=value,
                    sequence=ti.get("sequence", 0),
                    double_spend=ti.get("doubleSpentTxID", False),
                    strict=self.strict,
                )
        for to in tx["vout"]:
            value = int(round(float(to["value"]) * self.units, 0))
            t.add_output(
                value=value,
                lock_script=to["scriptPubKey"]["hex"],
                address=to["scriptPubKey"]["addresses"][0],
                spent=to["spent"],
                output_n=to["n"],
                strict=self.strict,
            )
        return t

    def getbalance(self, addresslist):
        balance = 0
        addresslist = self._addresslist_convert(addresslist)
        for a in addresslist:
            res = self.compose_request("address", a.address)
            balance += int(float(res["balance"]) / self.network.denominator)
        return balance

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        address = self._address_convert(address)
        res = self.compose_request("utxo", address.address)
        txs = []
        for tx in res:
            if tx["txid"] == after_txid:
                break
            txs.append(
                {
                    "address": address.address_orig,
                    "txid": tx["txid"],
                    "confirmations": tx["confirmations"],
                    "output_n": tx["vout"],
                    "input_n": 0,
                    "block_height": tx["height"],
                    "fee": None,
                    "size": 0,
                    "value": tx["satoshis"],
                    "script": tx["scriptPubKey"],
                    "date": None,
                }
            )
        return txs[::-1][:limit]

    def gettransaction(self, txid):
        tx = self.compose_request("tx", txid)
        return self._convert_to_transaction(tx)

    # FIXME: Not available anymore
    # def gettransactions(self, address, after_txid='', limit=MAX_TRANSACTIONS):
    #     address = self._address_convert(address)
    #     res = self.compose_request('addrs', address.address, 'txs')
    #     txs = []
    #     txs_dict = res['items'][::-1]
    #     if after_txid:
    #         txs_dict = txs_dict[[t['txid'] for t in txs_dict].index(after_txid) + 1:]
    #     for tx in txs_dict[:limit]:
    #         if tx['txid'] == after_txid:
    #             break
    #         txs.append(self._convert_to_transaction(tx))
    #     return txs

    def getrawtransaction(self, txid):
        res = self.compose_request("tx", txid)
        return res["hex"]

    def sendrawtransaction(self, rawtx):
        res = self.compose_request("sendtx", data=rawtx)
        return {"txid": res["result"], "response_dict": res}

    def estimatefee(self, blocks):
        res = self.compose_request("estimatefee", str(blocks))
        return int(float(res["result"]) / self.network.denominator)

    def blockcount(self):
        res = self.compose_request("status", "", variables={"q": "getinfo"})
        return res["blockbook"]["bestHeight"]

    def mempool(self, txid):
        res = self.compose_request("tx", txid)
        if res["confirmations"] == 0:
            return res["txid"]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        bd = self.compose_request("block", str(blockid))
        if parse_transactions:
            txs = []
            for tx in bd["txs"][(page - 1) * limit : page * limit]:
                # try:
                txs.append(self.gettransaction(tx["txid"]))
                # except Exception as e:
                #     _logger.error("Could not parse tx %s with error %s" % (tx['id'], e))
        else:
            txs = [tx["txid"] for tx in bd["txs"]]

        block = {
            "bits": bd["bits"],
            "depth": bd["confirmations"],
            "block_hash": bd["hash"],
            "height": bd["height"],
            "merkle_root": bd["merkleRoot"],
            "nonce": bd["nonce"],
            "prev_block": bd["previousBlockHash"],
            "time": bd["time"],
            "tx_count": len(bd["txs"]),
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None
            if not limit
            else int(len(bd["txs"]) // limit) + (len(bd["txs"]) % limit > 0),
            "limit": limit,
        }
        return block

    # def getrawblock(self, blockid):

    def isspent(self, txid, output_n):
        t = self.gettransaction(txid)
        return 1 if t.outputs[output_n].spent else 0

    def getinfo(self):
        info = self.compose_request("status", "")
        return {
            "blockcount": info["backend"]["blocks"],
            "chain": info["backend"]["chain"],
            "difficulty": int(float(info["backend"]["difficulty"])),
            "hashrate": 0,
            "mempool_size": info["blockbook"]["mempoolSize"],
        }



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    mempool.space client
#    © 2021-2022 - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.encoding import varstr
from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "mempool"
# Please note: In the Blockstream API, the first couple of Bitcoin blocks are not correctly indexed,
# so transactions from these blocks are missing.

_logger = logging.getLogger(__name__)


class MempoolClient(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self,
        function,
        data="",
        parameter="",
        parameter2="",
        variables=None,
        post_data="",
        method="get",
    ):
        url_path = function
        if data:
            url_path += "/" + data
        if parameter:
            url_path += "/" + parameter
        if parameter2:
            url_path += "/" + parameter2
        if variables is None:
            variables = {}
        if self.api_key:
            variables.update({"token": self.api_key})
        return self.request(url_path, variables, method, post_data=post_data)

    def getbalance(self, addresslist):
        balance = 0
        for address in addresslist:
            res = self.compose_request("address", address)
            balance += (
                res["chain_stats"]["funded_txo_sum"]
                - res["chain_stats"]["spent_txo_sum"]
            )
        return balance

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        self.latest_block = (
            self.blockcount() if not self.latest_block else self.latest_block
        )
        res = self.compose_request("address", address, "utxo")
        utxos = []
        # # key=lambda k: (k[2], pow(10, 20)-k[0].transaction_id, k[3]), reverse=True
        res = sorted(
            res,
            key=lambda k: 0
            if "block_height" not in k["status"]
            else k["status"]["block_height"],
        )
        for a in res:
            confirmations = 0
            block_height = None
            if "block_height" in a["status"]:
                block_height = a["status"]["block_height"]
                confirmations = self.latest_block - block_height
            utxos.append(
                {
                    "address": address,
                    "txid": a["txid"],
                    "confirmations": confirmations,
                    "output_n": a["vout"],
                    "input_n": 0,
                    "block_height": block_height,
                    "fee": None,
                    "size": 0,
                    "value": a["value"],
                    "script": "",
                    "date": None
                    if "block_time" not in a["status"]
                    else datetime.utcfromtimestamp(a["status"]["block_time"]),
                }
            )
            if a["txid"] == after_txid:
                utxos = []
        return utxos[:limit]

    def _parse_transaction(self, tx):
        block_height = (
            None if "block_height" not in tx["status"] else tx["status"]["block_height"]
        )
        confirmations = 0
        tx_date = None
        status = "unconfirmed"
        if tx["status"]["confirmed"]:
            if block_height:
                self.latest_block = (
                    self.blockcount() if not self.latest_block else self.latest_block
                )
                confirmations = self.latest_block - block_height + 1
            tx_date = datetime.utcfromtimestamp(tx["status"]["block_time"])
            status = "confirmed"

        t = BitcoinTransaction(
            locktime=tx["locktime"],
            version=tx["version"],
            network=self.network,
            block_height=block_height,
            fee=tx["fee"],
            size=tx["size"],
            txid=tx["txid"],
            date=tx_date,
            confirmations=confirmations,
            status=status,
            coinbase=tx["vin"][0]["is_coinbase"],
        )
        for ti in tx["vin"]:
            if ti["is_coinbase"]:
                t.add_input(
                    prev_txid=ti["txid"],
                    output_n=ti["vout"],
                    unlocking_script=ti["scriptsig"],
                    value=0,
                    sequence=ti["sequence"],
                    strict=self.strict,
                )
            else:
                t.add_input(
                    prev_txid=ti["txid"],
                    output_n=ti["vout"],
                    unlocking_script=ti["scriptsig"],
                    value=ti["prevout"]["value"],
                    address=ti["prevout"].get("scriptpubkey_address", ""),
                    unlocking_script_unsigned=ti["prevout"]["scriptpubkey"],
                    sequence=ti["sequence"],
                    witnesses=None
                    if "witness" not in ti
                    else [bytes.fromhex(w) for w in ti["witness"]],
                    strict=self.strict,
                )
        for to in tx["vout"]:
            t.add_output(
                value=to["value"],
                address=to.get("scriptpubkey_address", ""),
                spent=None,
                lock_script=to["scriptpubkey"],
                strict=self.strict,
            )
        if "segwit" in [i.witness_type for i in t.inputs] or "p2sh-segwit" in [
            i.witness_type for i in t.inputs
        ]:
            t.witness_type = "segwit"
        t.update_totals()
        return t

    def gettransaction(self, txid):
        tx = self.compose_request("tx", txid)
        return self._parse_transaction(tx)

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        prtxs = []
        before_txid = ""
        while True:
            txs = self.compose_request("address", address, "txs/chain", before_txid)
            prtxs += txs
            if len(txs) == 25:
                before_txid = txs[-1:][0]["txid"]
            else:
                break
            if len(prtxs) > limit:
                break
        txs = []
        for tx in prtxs[::-1]:
            t = self._parse_transaction(tx)
            if t:
                txs.append(t)
            if t.txid == after_txid:
                txs = []
            if len(txs) > limit:
                break
        return txs[:limit]

    def getrawtransaction(self, txid):
        return self.compose_request("tx", txid, "hex")

    def sendrawtransaction(self, rawtx):
        return self.compose_request("tx", post_data=rawtx, method="post")

    def estimatefee(self, blocks):
        estimates = self.compose_request("v1/fees", "recommended")
        if blocks < 2:
            return estimates["fastestFee"] * 1000
        elif blocks < 4:
            return estimates["halfHourFee"] * 1000
        if blocks < 7:
            return estimates["hourFee"] * 1000
        else:
            return estimates["minimumFee"] * 1000

    def blockcount(self):
        res = self.compose_request("blocks", "tip", "height")
        return res

    def mempool(self, txid=""):
        txids = self.compose_request("mempool", "txids")
        if not txid:
            return txids
        if txid in txids:
            return [txid]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        if isinstance(blockid, int):
            blockid = self.compose_request("block-height", str(blockid))
        if (page == 1 and limit == 10) or limit > 25:
            limit = 25
        bd = self.compose_request("block", blockid)
        btxs = self.compose_request("block", blockid, "txs", str((page - 1) * limit))
        if parse_transactions:
            txs = []
            for tx in btxs[:limit]:
                txs.append(self._parse_transaction(tx))
        else:
            txs = [tx["txid"] for tx in btxs]

        block = {
            "bits": bd["bits"],
            "depth": None,
            "block_hash": bd["id"],
            "height": bd["height"],
            "merkle_root": bd["merkle_root"],
            "nonce": bd["nonce"],
            "prev_block": bd["previousblockhash"],
            "time": bd["timestamp"],
            "tx_count": bd["tx_count"],
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None
            if not limit
            else int(bd["tx_count"] // limit) + (bd["tx_count"] % limit > 0),
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        if isinstance(blockid, int):
            blockid = self.compose_request("block-height", str(blockid))
        return self.compose_request("block", blockid, "raw").hex()

    def isspent(self, txid, output_n):
        res = self.compose_request("tx", txid, "outspend", str(output_n))
        return 1 if res["spent"] else 0

    # def getinfo(self):



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    dashd deamon
#    © 2018-2022 Oct - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# You can connect to a dash testnet deamon by adding the following provider to providers.json
# "dashd.testnet": {
#   "provider": "dashd",
#   "network": "dash_testnet",
#   "client_class": "DashdClient",
#   "url": "http://user:password@server_url:19998",
#   "api_key": "",
#   "priority": 11,
#   "denominator": 100000000
# }

import configparser

from fluxwallet.main import *
from fluxwallet.services.authproxy import AuthServiceProxy
from fluxwallet.services.baseclient import BaseClient
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "dashd"

_logger = logging.getLogger(__name__)


class ConfigError(Exception):
    def __init__(self, msg=""):
        self.msg = msg
        _logger.info(msg)

    def __str__(self):
        return self.msg


class DashdClient(BaseClient):
    """
    Class to interact with dashd, the Dash deamon
    """

    @staticmethod
    def from_config(configfile=None, network="dash"):
        """
        Read settings from dashd config file

        :param configfile: Path to config file. Leave empty to look in default places
        :type: str
        :param network: Dash mainnet or testnet. Default is dash mainnet
        :type: str

        :return DashdClient:
        """
        config = configparser.ConfigParser(strict=False)
        if not configfile:
            cfn = os.path.join(os.path.expanduser("~"), ".fluxwallet/dash.conf")
            if not os.path.isfile(cfn):
                cfn = os.path.join(os.path.expanduser("~"), ".dashcore/dash.conf")
            if not os.path.isfile(cfn):
                raise ConfigError(
                    "Please install dash client and specify a path to config file if path is not "
                    "default. Or place a config file in .fluxwallet/dash.conf to reference to "
                    "an external server."
                )
        else:
            cfn = os.path.join(FW_DATA_DIR, "config", configfile)
            if not os.path.isfile(cfn):
                raise ConfigError("Config file %s not found" % cfn)
        with open(cfn, "r") as f:
            config_string = "[rpc]\n" + f.read()
        config.read_string(config_string)

        try:
            if int(config.get("rpc", "testnet")):
                network = "testnet"
        except configparser.NoOptionError:
            pass
        if config.get("rpc", "rpcpassword") == "specify_rpc_password":
            raise ConfigError("Please update config settings in %s" % cfn)
        try:
            port = config.get("rpc", "port")
        except configparser.NoOptionError:
            if network == "testnet":
                port = 19998
            else:
                port = 9998
        server = "127.0.0.1"
        if "bind" in config["rpc"]:
            server = config.get("rpc", "bind")
        elif "externalip" in config["rpc"]:
            server = config.get("rpc", "externalip")
        url = "http://%s:%s@%s:%s" % (
            config.get("rpc", "rpcuser"),
            config.get("rpc", "rpcpassword"),
            server,
            port,
        )
        return DashdClient(network, url)

    def __init__(self, network="dash", base_url="", denominator=100000000, *args):
        """
        Open connection to dashcore node

        :param network: Dash mainnet or testnet. Default is dash mainnet
        :type: str
        :param base_url: Connection URL in format http(s)://user:password@host:port.
        :type: str
        :param denominator: Denominator for this currency. Should be always 100000000 (satoshis) for Dash
        :type: str
        """
        if not base_url:
            bdc = self.from_config("", network)
            base_url = bdc.base_url
            network = bdc.network
        if len(base_url.split(":")) != 4:
            raise ConfigError(
                "Dashd connection URL must be of format 'http(s)://user:password@host:port,"
                "current format is %s. Please set url in providers.json file" % base_url
            )
        if "password" in base_url:
            raise ConfigError(
                "Invalid password 'password' in dashd provider settings. "
                "Please set password and url in providers.json file"
            )
        _logger.info("Connect to dashd")
        self.proxy = AuthServiceProxy(base_url)
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def _parse_transaction(self, tx, block_height=None, get_input_values=True):
        t = BitcoinTransaction.parse_hex(
            tx["hex"], strict=self.strict, network=self.network
        )
        t.confirmations = None if "confirmations" not in tx else tx["confirmations"]
        if t.confirmations or block_height:
            t.status = "confirmed"
            t.verified = True
        for i in t.inputs:
            if i.prev_txid == b"\x00" * 32:
                i.script_type = "coinbase"
                continue
            if get_input_values:
                txi = self.proxy.getrawtransaction(i.prev_txid.hex(), 1)
                i.value = int(
                    round(
                        float(txi["vout"][i.output_n_int]["value"])
                        / self.network.denominator
                    )
                )
        for o in t.outputs:
            o.spent = None
        t.block_height = block_height
        t.version = tx["version"].to_bytes(4, "big")
        t.date = datetime.utcfromtimestamp(tx["blocktime"])
        t.update_totals()
        return t

    def gettransaction(self, txid):
        tx = self.proxy.getrawtransaction(txid, 1)
        return self._parse_transaction(tx)

    def getrawtransaction(self, txid):
        res = self.proxy.getrawtransaction(txid)
        return res

    def sendrawtransaction(self, rawtx):
        res = self.proxy.sendrawtransaction(rawtx)
        return {"txid": res, "response_dict": res}

    def estimatefee(self, blocks):
        try:
            res = self.proxy.estimatesmartfee(blocks)["feerate"]
        except KeyError:
            res = self.proxy.estimatefee(blocks)
        return int(res * self.units)

    def blockcount(self):
        return self.proxy.getblockcount()

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txs = []

        txs_list = self.proxy.listunspent(0, 99999999, [address])
        for t in sorted(txs_list, key=lambda x: x["confirmations"], reverse=True):
            txs.append(
                {
                    "address": t["address"],
                    "txid": t["txid"],
                    "confirmations": t["confirmations"],
                    "output_n": t["vout"],
                    "input_n": -1,
                    "block_height": None,
                    "fee": None,
                    "size": 0,
                    "value": int(t["amount"] * self.units),
                    "script": t["scriptPubKey"],
                    "date": None,
                }
            )
            if t["txid"] == after_txid:
                txs = []

        return txs

    def getblock(self, blockid, parse_transactions=True, page=1, limit=None):
        if isinstance(blockid, int):
            blockid = self.proxy.getblockhash(blockid)
        if not limit:
            limit = 99999

        txs = []
        if parse_transactions:
            bd = self.proxy.getblock(blockid, 2)
            for tx in bd["tx"][(page - 1) * limit : page * limit]:
                # try:
                tx["blocktime"] = bd["time"]
                tx["blockhash"] = bd["hash"]
                txs.append(
                    self._parse_transaction(
                        tx, block_height=bd["height"], get_input_values=False
                    )
                )
                # except Exception as e:
                #     _logger.error("Could not parse tx %s with error %s" % (tx['txid'], e))
            # txs += [tx['hash'] for tx in bd['tx'][len(txs):]]
        else:
            bd = self.proxy.getblock(blockid, 1)
            txs = bd["tx"]

        block = {
            "bits": int(bd["bits"], 16),
            "depth": bd["confirmations"],
            "hash": bd["hash"],
            "height": bd["height"],
            "merkle_root": bd["merkleroot"],
            "nonce": bd["nonce"],
            "prev_block": bd["previousblockhash"],
            "time": bd["time"],
            "total_txs": bd["nTx"],
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None,
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        if isinstance(blockid, int):
            blockid = self.proxy.getblockhash(blockid)
        return self.proxy.getblock(blockid, 0)

    def isspent(self, txid, index):
        res = self.proxy.gettxout(txid, index)
        if not res:
            return 1
        return 0

    def getinfo(self):
        info = self.proxy.getmininginfo()
        return {
            "blockcount": info["blocks"],
            "chain": info["chain"],
            "difficulty": int(info["difficulty"]),
            "hashrate": int(info["networkhashps"]),
            "mempool_size": int(info["pooledtx"]),
        }


if __name__ == "__main__":
    #
    # SOME EXAMPLES
    #

    from pprint import pprint

    # 1. Connect by specifying connection URL
    # base_url = 'http://dashrpcuser:passwd@host:9998'
    # bdc = DashdClient(base_url=base_url)
    # 2. Or connect using default settings or settings from config file
    bdc = DashdClient()

    print("\n=== SERVERINFO ===")
    pprint(bdc.proxy.getnetworkinfo())

    print("\n=== Best Block ===")
    blockhash = bdc.proxy.getbestblockhash()
    bestblock = bdc.proxy.getblock(blockhash)
    bestblock["tx"] = "..." + str(len(bestblock["tx"])) + " transactions..."
    pprint(bestblock)

    print("\n=== Mempool ===")
    rmp = bdc.proxy.getrawmempool()
    pprint(rmp[:25])
    print("... truncated ...")
    print("Mempool Size %d" % len(rmp))

    print("\n=== Raw BitcoinTransaction by txid ===")
    t = bdc.getrawtransaction(
        "c3d2a934ef8eb9b2291d113b330b9244c1521ef73df0a4b04c39e851112f01af"
    )
    pprint(t)

    print("\n=== Current network fees ===")
    t = bdc.estimatefee(5)
    pprint(t)



from .errors import WalletError

# wallet_transaction must be before wallet (Circular)
from .wallet_transaction import GenericTransaction, WalletTransaction
from .wallet import Wallet
from .wallet_key import WalletKey



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Cache DataBase - SqlAlchemy database definitions for caching
#    © 2020 February - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# try:
#     import mysql.connector
#     from parameterized import parameterized_class
#     import psycopg2
#     from psycopg2 import sql
#     from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
# except ImportError as e:
#     print("Could not import all modules. Error: %s" % e)
#     # from psycopg2cffi import compat  # Use for PyPy support
#     # compat.register()
#     pass  # Only necessary when mysql or postgres is used
from __future__ import annotations

from urllib.parse import urlparse

from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    LargeBinary,
    String,
    text,
)
from sqlalchemy.event import listen
from sqlalchemy.ext.asyncio import (
    AsyncAttrs,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.ext.asyncio.engine import AsyncEngine

# from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    WriteOnlyMapped,
    mapped_column,
    relationship,
)

from fluxwallet.main import *

# _logger = logging.getLogger(__name__)


class WitnessTypeTransactions(enum.Enum):
    legacy = "legacy"
    segwit = "segwit"


class Base(AsyncAttrs, DeclarativeBase):
    pass


class DbCache:
    """
    Cache Database object. Initialize database and open session when creating database object.

    Create new database if is doesn't exist yet

    """

    _built = False

    @classmethod
    async def start(cls, db_uri: str | None = None) -> DbCache:
        self = DbCache()

        async with self.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        await self.test_connection()
        DbCache._built = True

        return self

    async def __aenter__(self) -> AsyncSession:
        if not DbCache._built:
            async with self.engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)

            await self.test_connection()
            DbCache._built = True

        return self._session

    async def __aexit__(self, exception_type, exception_value, exception_traceback):
        # await self._session.close()
        # return None / False will reraise
        # print("Exiting DbCache Context")
        ...

    def __init__(
        self,
        db_uri: str | None = None,
        *,
        sessionmaker: async_sessionmaker | None = None,
    ):
        if db_uri is None:
            db_uri = DEFAULT_DATABASE_CACHE

        db_uri = f"sqlite+aiosqlite:///{db_uri}"

        self.connection_possible = False
        self.engine = create_async_engine(db_uri, isolation_level="READ UNCOMMITTED")
        self.sessionmaker = async_sessionmaker(self.engine, expire_on_commit=False)

        listen(self.engine.sync_engine, "connect", self.set_sqlite_pragma)
        self.db_uri = db_uri
        self._session = self.sessionmaker()

    def get_session(self):
        return self.sessionmaker()

    def set_sqlite_pragma(self, dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA journal_mode=WAL")
        cursor.close()

        # print("Cache Engine connected")

    async def test_connection(self) -> None:
        async with self.get_session() as session:
            try:
                await session.execute(text("SELECT 1"))
                self.connection_possible = True
            except Exception:
                self.connection_possible = False

    @property
    def session(self) -> AsyncSession:
        return self._session

    # def drop_db(self):
    #     self.session.commit()
    #     # self.session.close_all()
    #     close_all_sessions()
    #     Base.metadata.drop_all(self.engine)


class DbCacheTransactionNode(Base):
    """
    Link table for cache transactions and addresses
    """

    __tablename__ = "cache_transactions_node"
    txid = Column(
        LargeBinary(32), ForeignKey("cache_transactions.txid"), primary_key=True
    )
    transaction = relationship(
        "DbCacheTransaction", back_populates="nodes", doc="Related transaction object"
    )
    index_n = Column(
        Integer, primary_key=True, doc="Order of input/output in this transaction"
    )
    value = Column(BigInteger, default=0, doc="Value of transaction input")
    address = Column(
        String(255), index=True, doc="Address string base32 or base58 encoded"
    )
    script = Column(LargeBinary, doc="Locking or unlocking script")
    witnesses = Column(
        LargeBinary, doc="Witnesses (signatures) used in Segwit transaction inputs"
    )
    sequence = Column(
        BigInteger,
        default=0xFFFFFFFF,
        doc="Transaction sequence number. Used for timelock transaction inputs",
    )
    is_input = Column(Boolean, primary_key=True, doc="True if input, False if output")
    spent = Column(Boolean, default=None, doc="Is output spent?")
    ref_txid = Column(
        LargeBinary(32),
        index=True,
        doc="Transaction hash of input which spends this output",
    )
    ref_index_n = Column(
        BigInteger, doc="Index number of transaction input which spends this output"
    )

    def prev_txid(self):
        if self.is_input:
            return self.ref_txid

    def output_n(self):
        if self.is_input:
            return self.ref_index_n

    def spending_txid(self):
        if not self.is_input:
            return self.ref_txid

    def spending_index_n(self):
        if not self.is_input:
            return self.ref_index_n


class DbCacheTransaction(Base):
    """
    Transaction Cache Table

    Database which stores transactions received from service providers as cache

    """

    __tablename__ = "cache_transactions"
    txid = Column(
        LargeBinary(32),
        primary_key=True,
        doc="Hexadecimal representation of transaction hash or transaction ID",
    )
    date = Column(
        DateTime, doc="Date when transaction was confirmed and included in a block"
    )
    version = Column(
        BigInteger,
        default=1,
        doc="Tranaction version. Default is 1 but some wallets use another version number",
    )
    locktime = Column(
        BigInteger,
        default=0,
        doc="Transaction level locktime. Locks the transaction until a specified block "
        "(value from 1 to 5 million) or until a certain time (Timestamp in seconds after 1-jan-1970)."
        " Default value is 0 for transactions without locktime",
    )
    expiry_height = Column(
        Integer,
        default=0,
        doc="Expiry height (in blocktime) for this transactions. I.e, must be confirmed before this block height."
        "Default is 0: no expiry (not version 4 transaction)",
    )
    confirmations = Column(
        Integer,
        default=0,
        doc="Number of confirmation when this transaction is included in a block. "
        "Default is 0: unconfirmed",
    )
    block_height = Column(
        Integer, index=True, doc="Height of block this transaction is included in"
    )
    network_name = Column(String(20), doc="Blockchain network name of this transaction")
    fee = Column(BigInteger, doc="Transaction fee")
    nodes = relationship(
        "DbCacheTransactionNode",
        cascade="all,delete",
        doc="List of all inputs and outputs as DbCacheTransactionNode objects",
    )
    order_n = Column(Integer, doc="Order of transaction in block")
    witness_type = Column(
        Enum(WitnessTypeTransactions),
        default=WitnessTypeTransactions.legacy,
        doc="Transaction type enum: legacy or segwit",
    )


class DbCacheAddress(Base):
    """
    Address Cache Table

    Stores transactions and unspent outputs (UTXO's) per address

    """

    __tablename__ = "cache_address"
    address = Column(
        String(255), primary_key=True, doc="Address string base32 or base58 encoded"
    )
    network_name = Column(String(20), doc="Blockchain network name of this transaction")
    balance = Column(
        BigInteger, default=0, doc="Total balance of UTXO's linked to this key"
    )
    last_block = Column(Integer, doc="Number of last updated block")
    last_txid = Column(
        LargeBinary(32), doc="Transaction ID of latest transaction in cache"
    )
    last_tx_index = Column(
        Integer, doc="Transaction index of latest transaction in cache"
    )
    n_utxos = Column(Integer, doc="Total number of UTXO's for this address")
    n_txs = Column(Integer, doc="Total number of transactions for this address")


class DbCacheBlock(Base):
    """
    Block Cache Table

    Stores block headers
    """

    __tablename__ = "cache_blocks"
    height = Column(
        Integer, primary_key=True, doc="Height or sequence number for this block"
    )
    block_hash = Column(LargeBinary(32), index=True, doc="Hash of this block")
    network_name = Column(String(20), doc="Blockchain network name")
    version = Column(
        BigInteger, doc="Block version to specify which features are used (hex)"
    )
    prev_block = Column(LargeBinary(32), doc="Block hash of previous block")
    merkle_root = Column(
        LargeBinary(32), doc="Merkle root used to validate transaction in block"
    )
    time = Column(BigInteger, doc="Timestamp to indicated when block was created")
    bits = Column(
        BigInteger,
        doc="Encoding for proof-of-work, used to determine target and difficulty",
    )
    nonce = Column(
        BigInteger,
        doc="Nonce (number used only once or n-once) is used to create different block hashes",
    )
    tx_count = Column(Integer, doc="Number of transactions included in this block")


class DbCacheVars(Base):
    """
    Table to store various blockchain related variables
    """

    __tablename__ = "cache_variables"
    varname = Column(String(50), primary_key=True, doc="Variable unique name")
    network_name = Column(
        String(20), primary_key=True, doc="Blockchain network name of this transaction"
    )
    value = Column(String(255), doc="Value of variable")
    type = Column(String(20), doc="Type of variable: int, string or float")
    expires = Column(DateTime, doc="Datetime value when variable expires")



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    © 2017 May - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import fluxwallet.services.authproxy
import fluxwallet.services.baseclient
import fluxwallet.services.bcoin
import fluxwallet.services.bitaps
import fluxwallet.services.bitcoind
import fluxwallet.services.bitcoinlibtest
import fluxwallet.services.bitflyer
import fluxwallet.services.bitgo
import fluxwallet.services.blockchaininfo
import fluxwallet.services.blockchair
import fluxwallet.services.blockcypher
import fluxwallet.services.blocksmurfer
import fluxwallet.services.blockstream
import fluxwallet.services.chainso
import fluxwallet.services.cryptoid
import fluxwallet.services.dashd
import fluxwallet.services.dogecoind
import fluxwallet.services.insightdash
import fluxwallet.services.litecoinblockexplorer
import fluxwallet.services.litecoind
import fluxwallet.services.litecoreio
import fluxwallet.services.mempool
import fluxwallet.services.runonflux



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    CryptoID Chainz client
#    © 2018-2022 October - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

_logger = logging.getLogger(__name__)

PROVIDERNAME = "cryptoid"


class CryptoID(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(self, func=None, path_type="api", variables=None, method="get"):
        # API path: http://chainz.cryptoid.info/ltc/api.dws
        # Explorer path for raw tx: https://chainz.cryptoid.info/explorer/tx.raw.dws
        if variables is None:
            variables = {}
        if path_type == "api":
            url_path = "%s/api.dws" % self.provider_coin_id
            variables.update({"q": func})
        else:
            url_path = "explorer/tx.raw.dws"
            variables.update({"coin": self.provider_coin_id})
        # if not self.api_key:
        #     raise ClientError("Request a CryptoID API key before using this provider")
        if self.api_key:
            variables.update({"key": self.api_key})
        return self.request(url_path, variables, method)

    def getbalance(self, addresslist):
        balance = 0.0
        addresslist = self._addresslist_convert(addresslist)
        for a in addresslist:
            res = self.compose_request("getbalance", variables={"a": a.address})
            balance += float(res)
        return int(balance * self.units)

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        if not self.api_key:
            raise ClientError(
                "Method getutxos() is not available for CryptoID without API key"
            )
        utxos = []
        address = self._address_convert(address)
        variables = {"active": address.address}
        res = self.compose_request("unspent", variables=variables)
        if len(res["unspent_outputs"]) > 50:
            _logger.info(
                "CryptoID: Large number of outputs for address %s, "
                "UTXO list may be incomplete" % address.address
            )
        for utxo in res["unspent_outputs"][::-1]:
            if utxo["txid"] == after_txid:
                break
            utxos.append(
                {
                    "address": address.address_orig,
                    "txid": utxo["txid"],
                    "confirmations": utxo["confirmations"],
                    "output_n": utxo["tx_output_n"]
                    if "tx_output_n" in utxo
                    else utxo["tx_ouput_n"],
                    "input_n": 0,
                    "block_height": None,
                    "fee": None,
                    "size": 0,
                    "value": int(utxo["value"]),
                    "script": utxo["script"],
                    "date": None,
                }
            )
        return utxos[::-1][:limit]

    def gettransaction(self, txid):
        variables = {"id": txid, "hex": None}
        tx = self.compose_request(path_type="explorer", variables=variables)
        t = BitcoinTransaction.parse_hex(
            tx["hex"], strict=self.strict, network=self.network
        )
        variables = {"t": txid}
        tx_api = self.compose_request("txinfo", path_type="api", variables=variables)
        for n, i in enumerate(t.inputs):
            if i.script_type != "coinbase":
                i.value = int(round(tx_api["inputs"][n]["amount"] * self.units, 0))
            else:
                i.value = 0
                t.coinbase = True
        for n, o in enumerate(t.outputs):
            o.spent = None
        if tx["confirmations"]:
            t.status = "confirmed"
        else:
            t.status = "unconfirmed"
        t.date = datetime.utcfromtimestamp(tx["time"])
        t.block_height = tx_api["block"]
        t.block_hash = tx["blockhash"]
        t.confirmations = tx["confirmations"]
        t.rawtx = bytes.fromhex(tx["hex"])
        t.size = tx["size"]
        t.network = self.network
        t.locktime = tx["locktime"]
        t.version = tx["version"].to_bytes(4, "big")
        t.output_total = int(round(tx_api["total_output"] * self.units, 0))
        t.input_total = int(round(tx_api["total_input"] * self.units, 0))
        t.fee = 0
        if t.input_total:
            t.fee = t.input_total - t.output_total
        return t

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        address = self._address_convert(address)
        txs = []
        txids = []
        variables = {"active": address.address, "n": 100}
        res = self.compose_request("multiaddr", variables=variables)
        for tx in res["txs"]:
            if tx["hash"] not in txids:
                txids.insert(0, tx["hash"])
        if after_txid:
            txids = txids[txids.index(after_txid) + 1 :]
        for txid in txids[:limit]:
            t = self.gettransaction(txid)
            txs.append(t)
        return txs

    def getrawtransaction(self, txid):
        variables = {"id": txid, "hex": None}
        tx = self.compose_request(path_type="explorer", variables=variables)
        return tx["hex"]

    # def sendrawtransaction

    # def estimatefee

    def blockcount(self):
        r = self.compose_request("getblockcount", path_type="api")
        return r

    def mempool(self, txid):
        variables = {"id": txid, "hex": None}
        tx = self.compose_request(path_type="explorer", variables=variables)
        if "confirmations" not in tx:
            return [tx["txid"]]
        return []

    # def getblock

    # def isspent

    # def getinfo(self):



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    SERVICES - Main Service connector
#    © 2017-2022 - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fluxwallet.wallet import GenericTransaction

import asyncio
import inspect
import json
import logging
import random
import time
from asyncio import Queue
from collections.abc import Iterator
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path

from rich.pretty import pprint
from sqlalchemy import func, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from fluxwallet import services

# from fluxwallet.blocks import Block
from fluxwallet.config.config import (
    BLOCK_COUNT_CACHE_TIME,
    DEFAULT_NETWORK,
    FW_DATA_DIR,
    MAX_TRANSACTIONS,
    SERVICE_CACHING_ENABLED,
    SERVICE_MAX_ERRORS,
    TIMEOUT_REQUESTS,
    TYPE_TEXT,
)
from fluxwallet.db_cache import (
    DbCache,
    DbCacheAddress,
    DbCacheBlock,
    DbCacheTransaction,
    DbCacheTransactionNode,
    DbCacheVars,
)
from fluxwallet.encoding import int_to_varbyteint, to_bytes, varbyteint_to_int, varstr
from fluxwallet.networks import Network
from fluxwallet.transactions import (
    BitcoinTransaction,
    FluxTransaction,
    transaction_update_spents,
)

_logger = logging.getLogger(__name__)


def test_cache(f):
    @wraps(f)
    async def inner(ref: Cache, *args, **kwargs):
        if await ref.cache_enabled():
            return await f(ref, *args, **kwargs)

    return inner


class SingletonNetwork(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        network = kwargs["network"]

        if network not in cls._instances:
            cls._instances[network] = super().__call__(*args, **kwargs)

        return cls._instances[network]


class ServiceError(Exception):
    def __init__(self, msg=""):
        self.msg = msg
        _logger.error(msg)

    def __str__(self):
        return self.msg


class Service(metaclass=SingletonNetwork):
    """
    Class to connect to various cryptocurrency service providers. Use to receive network and blockchain information,
    get specific transaction information, current network fees or push a raw transaction.

    The Service class connects to 1 or more service providers at random to retrieve or send information. If a service
    providers fails to correctly respond the Service class will try another available provider.

    """

    def __init__(
        self,
        *,
        network: str | Network = DEFAULT_NETWORK,
        min_providers: int = 1,
        max_providers: int = 1,
        providers: list | None = None,
        timeout: int = TIMEOUT_REQUESTS,
        cache_uri: str | None = None,
        ignore_priority: bool = False,
        exclude_providers: list | None = None,
        max_errors: int = SERVICE_MAX_ERRORS,
        strict: bool = True,
    ):
        """
        Create a service object for the specified network. By default, the object connect to 1 service provider, but you
        can specify a list of providers or a minimum or maximum number of providers.

        :param network: Specify network used
        :type network: str, Network
        :param min_providers: Minimum number of providers to connect to. Default is 1. Use for instance to receive fee information from a number of providers and calculate the average fee.
        :type min_providers: int
        :param max_providers: Maximum number of providers to connect to. Default is 1.
        :type max_providers: int
        :param providers: List of providers to connect to. Default is all providers and select a provider at random.
        :type providers: list of str
        :param timeout: Timeout for web requests. Leave empty to use default from config settings
        :type timeout: int
        :param cache_uri: Database to use for caching
        :type cache_uri: str
        :param ignore_priority: Ignores provider priority if set to True. Could be used for unit testing, so no providers are missed when testing. Default is False
        :type ignore_priority: bool
        :param exclude_providers: Exclude providers in this list, can be used when problems with certain providers arise.
        :type exclude_providers: list of str
        :param strict: Strict checks of valid signatures, scripts and transactions. Normally use strict=True for wallets, transaction validations etcetera. For blockchain parsing strict=False should be used, but be sure to check warnings in the log file. Default is True.
        :type strict: bool
        """

        if not isinstance(network, Network):
            self.network = Network(network)
        else:
            self.network = network

        if min_providers > max_providers:
            max_providers = min_providers

        fn = Path(FW_DATA_DIR, "providers.json")
        f = fn.open("r")

        try:
            self.providers_defined = json.loads(f.read())
        except json.decoder.JSONDecodeError as e:  # pragma: no cover
            errstr = "Error reading provider definitions from %s: %s" % (fn, e)
            _logger.warning(errstr)
            raise ServiceError(errstr)
        f.close()

        provider_list = list(
            [self.providers_defined[x]["provider"] for x in self.providers_defined]
        )
        if providers is None:
            providers = []
        if exclude_providers is None:
            exclude_providers = []
        if not isinstance(providers, list):
            providers = [providers]
        for p in providers:
            if p not in provider_list:
                raise ServiceError(
                    "Provider '%s' not found in provider definitions" % p
                )

        self.providers = {}
        for p in self.providers_defined:
            if (
                self.providers_defined[p]["network"] == network
                or self.providers_defined[p]["network"] == ""
            ) and (not providers or self.providers_defined[p]["provider"] in providers):
                self.providers.update({p: self.providers_defined[p]})
        for nop in exclude_providers:
            if nop in self.providers:
                del self.providers[nop]

        if not self.providers:
            raise ServiceError("No providers found for network %s" % network)
        self.min_providers = min_providers
        self.max_providers = max_providers
        self.results = {}
        self.errors = {}
        self.resultcount = 0
        self.max_errors = max_errors
        self.complete = None
        self.timeout = timeout
        self._blockcount_update = 0
        self._blockcount = 0
        self.cache = None
        self.cache_uri = cache_uri

        try:
            self.cache = Cache(self.network, db_uri=cache_uri)
        except Exception as e:
            self.cache = Cache(self.network, db_uri="")
            _logger.warning("Could not connect to cache database. Error: %s" % e)

        self.results_cache_n = 0
        self.ignore_priority = ignore_priority
        self.strict = strict
        # if self.min_providers > 1:
        #     self._blockcount = Service(
        #         network=network, cache_uri=cache_uri
        #     ).blockcount()
        # else:
        #     self._blockcount = self.blockcount()

    # def __exit__(self):
    #     try:
    #         self.cache.session.close()
    #     except Exception:
    #         pass

    def _reset_results(self):
        self.results = {}
        self.errors = {}
        self.complete = None
        self.resultcount = 0

    async def _provider_execute(self, method: str, *arguments):
        self._reset_results()

        provider_lst = [
            p[0]
            for p in sorted(
                [(x, self.providers[x]["priority"]) for x in self.providers],
                key=lambda x: (x[1], random.random()),
                reverse=True,
            )
        ]

        if self.ignore_priority:
            random.shuffle(provider_lst)

        for sp in provider_lst:
            if self.resultcount >= self.max_providers:
                break
            try:
                if (
                    sp not in ["bitcoind", "litecoind", "dashd", "dogecoind", "caching"]
                    and not self.providers[sp]["url"]
                    and self.network.name != "fluxwallet_test"
                ):
                    continue
                client = getattr(services, self.providers[sp]["provider"])
                providerclient = getattr(client, self.providers[sp]["client_class"])
                pc_instance = providerclient(
                    self.network,
                    self.providers[sp]["url"],
                    self.providers[sp]["denominator"],
                    self.providers[sp]["api_key"],
                    self.providers[sp]["provider_coin_id"],
                    self.providers[sp]["network_overrides"],
                    self.timeout,
                    self._blockcount,
                    self.strict,
                )
                if not hasattr(pc_instance, method):
                    _logger.debug("Method %s not found for provider %s" % (method, sp))
                    continue
                if self.providers[sp]["api_key"] == "api-key-needed":
                    _logger.debug("API key needed for provider %s" % sp)
                    continue
                providermethod = getattr(pc_instance, method)

                if asyncio.iscoroutinefunction(providermethod):
                    res = await providermethod(*arguments)
                else:
                    # this can be an Iterator
                    res = providermethod(*arguments)

                if res is False:  # pragma: no cover
                    self.errors.update({sp: "Received empty response"})
                    _logger.info(
                        "Empty response from %s when calling %s" % (sp, method)
                    )
                    continue
                self.results.update({sp: res})
                _logger.debug("Executed method %s from provider %s" % (method, sp))
                self.resultcount += 1
            except Exception as e:
                # remember to fix this!
                raise

                if not isinstance(e, AttributeError):
                    try:
                        err = e.msg
                    except AttributeError:
                        err = e
                    self.errors.update({sp: err})
                    _logger.debug("Error %s on provider %s" % (e, sp))
                    # -- Use this to debug specific Services errors --
                    # from pprint import pprint
                    # pprint(self.errors)

                if len(self.errors) >= self.max_errors:
                    _logger.warning(
                        "Aborting, max errors exceeded: %s" % list(self.errors.keys())
                    )
                    if len(self.results):
                        return list(self.results.values())[0]
                    else:
                        return False

            if self.resultcount >= self.max_providers:
                break

        if not self.resultcount:
            raise ServiceError(
                "No successful response from any serviceprovider: %s"
                % list(self.providers.keys())
            )

        # ok
        return list(self.results.values())[0]

    async def getbalance(
        self, addresslist: list[str] | str, addresses_per_request: int = 5
    ) -> int:
        """
        Get total balance for address or list of addresses

        :param addresslist: Address or list of addresses
        :type addresslist: list, str
        :param addresses_per_request: Maximum number of addresses per request. Default is 5. Use lower setting when you experience timeouts or service request errors, or higher when possible.
        :type addresses_per_request: int

        :return dict: Balance per address
        """
        if isinstance(addresslist, TYPE_TEXT):
            addresslist = [addresslist]

        tot_balance = 0
        while addresslist:
            for address in addresslist:
                db_addr = await self.cache.getaddress(address)
                if (
                    db_addr
                    and db_addr.last_block
                    and db_addr.last_block >= await self.blockcount()
                    and db_addr.balance
                ):
                    tot_balance += db_addr.balance
                    addresslist.remove(address)

            balance: int = await self._provider_execute(
                "getbalance", addresslist[:addresses_per_request]
            )
            if balance:
                tot_balance += balance

            if len(addresslist) == 1:
                await self.cache.store_address(addresslist[0], balance=balance)

            addresslist = addresslist[addresses_per_request:]
        return tot_balance

    async def getutxos(
        self, address: str, after_txid: str = "", limit: int = MAX_TRANSACTIONS
    ):
        """
        Get list of unspent outputs (UTXO's) for specified address.

        Sorted from old to new, so the highest number of confirmations first.

        :param address: Address string
        :type address: str
        :param after_txid: Transaction ID of last known transaction. Only check for utxos after given tx id. Default: Leave empty to return all utxos.
        :type after_txid: str
        :param limit: Maximum number of utxo's to return. Sometimes ignored by service providers if more results are returned by default.
        :type limit: int

        :return list: UTXO's for address
        """
        if not isinstance(address, TYPE_TEXT):
            raise ServiceError("Address parameter must be of type text")

        self.results_cache_n = 0
        self.complete = True

        utxos_cache = []
        if self.min_providers <= 1:
            utxos_cache = await self.cache.getutxos(address, bytes.fromhex(after_txid))

        if utxos_cache:
            self.results_cache_n = len(utxos_cache)

            # Last updated block does not always include spent info...
            # if db_addr and db_addr.last_block and db_addr.last_block >= self.blockcount():
            #     return utxos_cache
            after_txid = utxos_cache[-1:][0]["txid"]

        utxos: list[dict] = await self._provider_execute(
            "getutxos", address, after_txid, limit
        )
        # start raising errors

        # if utxos is False:
        #     raise ServiceError("Error when retrieving UTXO's")

        # Update cache_transactions_node
        tasks = []
        for utxo in utxos:
            tasks.append(self.cache.store_utxo(utxo["txid"], utxo["output_n"]))

        await asyncio.gather(*tasks)

        if utxos and len(utxos) >= limit:
            self.complete = False
        elif not after_txid:
            balance = sum(u["value"] for u in utxos)
            await self.cache.store_address(address, balance=balance, n_utxos=len(utxos))

        return utxos_cache + utxos

    async def get_transaction(self, txid: str) -> GenericTransaction:
        """
        Get a transaction by its transaction hash. Convert to fluxwallet Transaction object.

        :param txid: Transaction identification hash
        :type txid: str

        :return Transaction: A single transaction object
        """
        tx = None
        self.results_cache_n = 0

        if self.min_providers <= 1:
            tx = await self.cache.get_transaction(bytes.fromhex(txid))
            if tx:
                self.results_cache_n = 1

        if not tx:
            tx: GenericTransaction = await self._provider_execute(
                "get_transaction", txid
            )

            if tx and tx.txid != txid:
                _logger.warning("Incorrect txid after parsing")
                tx.txid = txid

            if len(self.results) and self.min_providers <= 1:
                await self.cache.store_transaction(tx)

        return tx

    async def get_transactions(self, txids: list) -> list[GenericTransaction]:
        """
        Get a transaction by its transaction hash. Convert to fluxwallet Transaction object.

        :param txid: Transaction identification hash
        :type txid: str

        :return Transaction: A single transaction object
        """

        # this is just for tests to be able to do assertions
        self.results_cache_n = 0

        cache_txs: list[GenericTransaction] = []
        provider_txs: list[GenericTransaction] = []
        cache_misses: list[str] = []

        if self.min_providers <= 1:
            for txid in txids:
                tx = await self.cache.get_transaction(bytes.fromhex(txid))
                if tx:
                    self.results_cache_n += 1
                    cache_txs.append(tx)
                else:
                    cache_misses.append(txid)

            txs: list[GenericTransaction] = await self._provider_execute(
                "get_transactions", cache_misses
            )

            for tx in txs:
                if tx and tx.txid not in cache_misses:
                    _logger.warning("Incorrect txid after parsing")
                    continue
                    # tx.txid = txid
                provider_txs.append(tx)

                # what is this results thing?
                if len(self.results):
                    await self.cache.store_transaction(tx)

        return cache_txs + provider_txs

    async def get_transactions_by_address(
        self, address: str, after_tx_index: int = 0, limit: int = MAX_TRANSACTIONS
    ) -> Iterator[list[FluxTransaction]]:
        """
        Get all transactions for specified address.

        Sorted from old to new, so transactions with highest number of confirmations first.

        :param address: Address string
        :type address: str
        :param after_txid: Transaction ID of last known transaction. Only check for transactions after given tx id. Default: Leave empty to return all transaction. If used only provide a single address
        :type after_txid: str
        :param limit: Maximum number of transactions to return
        :type limit: int

        :return list: List of Transaction objects
        """
        self._reset_results()
        self.results_cache_n = 0

        # if not address:
        #     yield []

        if not isinstance(address, TYPE_TEXT):
            raise ServiceError("Address parameter must be of type text")

        db_addr = await self.cache.getaddress(address)
        txs_cache: list[GenericTransaction] = []
        # qry_after_txid = bytes.fromhex(after_txid)

        # Retrieve transactions from cache
        caching_enabled = True
        if self.min_providers > 1:  # Disable cache if comparing providers
            caching_enabled = False

        if caching_enabled:
            # change this to yield
            if txs_cache := await self.cache.get_transactions(
                address, after_tx_index, limit
            ):
                self.results_cache_n = len(txs_cache)
                # if len(txs_cache) == limit:
                #     return txs_cache

                # limit = limit - len(txs_cache)
                # qry_after_txid = bytes.fromhex(txs_cache[-1:][0].txid)

        if txs_cache:
            txs_cache = transaction_update_spents(txs_cache, address)
            yield txs_cache

        # Get (extra) transactions from service providers
        tx_generator = None

        # if our cache isn't up to date with the latest block, go out to service provider and get more
        if (
            not (
                db_addr
                and db_addr.last_block
                and db_addr.last_block >= await self.blockcount()
            )
            or not caching_enabled
        ):
            tx_generator = await self._provider_execute(
                "get_transactions_by_address", address, after_tx_index, limit
            )
            # start raising errors instead of False

            # if txs is False:
            #     raise ServiceError(
            #         "Error when retrieving transactions from service provider"
            #     )

        # Store transactions and address in cache
        # - disable cache if comparing providers or if after_txid is used and no cache is available
        last_block = await self.blockcount()
        # last_txid = qry_after_txid
        self.complete = True
        # store_in_cache = False
        # if (
        #     self.min_providers <= 1
        #     and not (after_txid and not db_addr)
        #     and caching_enabled
        # ):
        #     store_in_cache = True
        # if len(txs) == limit:
        #     self.complete = False
        #     last_block = txs[-1:][0].block_height

        if not tx_generator:
            # print(db_addr, db_addr.last_block, await self.blockcount())
            # empty generator
            # yield []
            return

        new_tx_count = 0
        async for txs in tx_generator:
            new_tx_count += len(txs)
            # if len(txs) and not last_txid:
            #     last_txid = bytes.fromhex(txs[0].txid)

            # order_n = 0
            # tasks = []
            # if store_in_cache:
            #     for t in txs:
            #         if t.confirmations != 0:
            #             # tasks.append(self.cache.store_transaction(t, order_n))
            #             tasks.append(self.cache.store_transaction(t))
            #             order_n += 1

            #     await asyncio.gather(*tasks)
            txs = transaction_update_spents(txs, address)

            if caching_enabled:
                # await self.cache.store_address(
                #     address, last_block, last_txid=last_txid, txs_complete=True
                # )

                # tasks = []
                for t in txs:
                    # asyncio.create_task(self.cache.store_transaction(t))
                    await self.cache.store_transaction(t)
                # await asyncio.gather(*tasks)

            yield txs

        if caching_enabled:
            await self.cache.store_address(
                address,
                last_block,
                last_tx_index=new_tx_count + after_tx_index,
                txs_complete=self.complete,
            )

    async def getrawtransaction(self, txid: str) -> str:
        """
        Get a raw transaction by its transaction hash

        :param txid: Transaction identification hash
        :type txid: str

        :return str: Raw transaction as hexstring
        """
        self.results_cache_n = 0
        rawtx = await self.cache.getrawtransaction(bytes.fromhex(txid))
        if rawtx:
            self.results_cache_n = 1
            return rawtx

        return await self._provider_execute("getrawtransaction", txid)

    async def sendrawtransaction(self, rawtx: str) -> dict:
        """
        Push a raw transaction to the network

        :param rawtx: Raw transaction as hexstring or bytes
        :type rawtx: str

        :return dict: Send transaction result
        """
        return await self._provider_execute("sendrawtransaction", rawtx)

    async def estimatefee(self, blocks: int = 3, priority: str = ""):
        """
        Estimate fee per kilobyte for a transaction for this network with expected confirmation within a certain
        amount of blocks

        :param blocks: Expected confirmation time in blocks. Default is 3.
        :type blocks: int
        :param priority: Priority for transaction: can be 'low', 'medium' or 'high'. Overwrites value supplied in 'blocks' argument
        :type priority: str

        :return int: Fee in the smallest network denominator (satoshi)
        """
        self.results_cache_n = 0

        if priority:
            if priority == "low":
                blocks = 10
            elif priority == "high":
                blocks = 1

        if self.min_providers <= 1:  # Disable cache if comparing providers
            fee = await self.cache.estimatefee(blocks)
            if fee:
                self.results_cache_n = 1
                return fee

        fee = await self._provider_execute("estimatefee", blocks)
        if not fee:  # pragma: no cover
            if self.network.fee_default:
                fee = self.network.fee_default
            else:
                raise ServiceError(
                    "Could not estimate fees, please define default fees in network settings"
                )
        if fee < self.network.fee_min:
            fee = self.network.fee_min
        elif fee > self.network.fee_max:
            fee = self.network.fee_max

        await self.cache.store_estimated_fee(blocks, fee)

        return fee

    async def blockcount(self) -> int:
        """
        Get the latest block number: The block number of last block in longest chain on the Blockchain.

        Block count is cashed for BLOCK_COUNT_CACHE_TIME seconds to avoid to many calls to service providers.

        :return int:
        """

        blockcount = await self.cache.blockcount()
        last_cache_blockcount = await self.cache.blockcount(never_expires=True)

        if blockcount:
            self._blockcount = blockcount
            return blockcount

        current_timestamp = time.time()
        if self._blockcount_update < current_timestamp - BLOCK_COUNT_CACHE_TIME:
            new_count = await self._provider_execute("blockcount")

            if not self._blockcount or (new_count and new_count > self._blockcount):
                self._blockcount = new_count
                self._blockcount_update = current_timestamp

            # if cache is disabled, returns None
            if last_cache_blockcount and last_cache_blockcount > self._blockcount:
                return last_cache_blockcount

            # Store result in cache
            if len(self.results) and list(self.results.keys())[0] != "caching":
                await self.cache.store_blockcount(self._blockcount)

        return self._blockcount

    async def getblock(
        self,
        blockid: int | str,
        parse_transactions: bool = True,
        page: int = 1,
        limit: int | None = None,
    ) -> Block | None:
        """
        Get block with specified block height or block hash from service providers.

        If parse_transaction is set to True a list of Transaction object will be returned otherwise a
        list of transaction ID's.

        Some providers require 1 or 2 extra request per transaction, so to avoid timeouts or rate limiting errors
        you can specify a page and limit for the transaction. For instance with page=2, limit=4 only transaction
        5 to 8 are returned to the Blocks's 'transaction' attribute.

        If you only use a local bcoin or bitcoind provider, make sure you set the limit to maximum (i.e. 9999)
        because all transactions are already downloaded when fetching the block.

        >>> from fluxwallet.services.services import Service
        >>> srv = Service()
        >>> b = srv.getblock(0)
        >>> b
        <Block(000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f, 0, transactions: 1)>

        :param blockid: Hash or block height of block
        :type blockid: str, int
        :param parse_transactions: Return Transaction objects or just transaction ID's. Default is return txids.
        :type parse_transactions: bool
        :param page: Page number of transaction paging. Default is start from the beginning: 1
        :type page: int
        :param limit: Maximum amount of transaction to return. Default is 25 if parse transaction is enabled, otherwise returns all txid's (9999)
        :type limit: int

        :return Block:
        """
        if limit is None:
            limit = 25 if parse_transactions else 99999

        block = await self.cache.getblock(blockid)
        is_last_page = False

        if block:
            # Block found get transactions from cache
            block.transactions = await self.cache.getblocktransactions(
                block.height, page, limit
            )
            if block.transactions:
                self.results_cache_n = 1
            is_last_page = page * limit > block.tx_count
        if (
            not block
            or (not len(block.transactions) and limit != 0)
            or (not is_last_page and len(block.transactions) < limit)
            or (
                is_last_page
                and ((page - 1) * limit - block.tx_count + len(block.transactions)) < 0
            )
        ):
            self.results_cache_n = 0
            bd = await self._provider_execute(
                "getblock", blockid, parse_transactions, page, limit
            )
            if not bd:
                return

            block = Block(
                bd["block_hash"],
                bd["version"],
                bd["prev_block"],
                bd["merkle_root"],
                bd["time"],
                bd["bits"],
                bd["nonce"],
                bd["txs"],
                bd["height"],
                bd["depth"],
                self.network,
            )
            block.tx_count = bd["tx_count"]
            block.limit = limit
            block.page = page

            if parse_transactions and self.min_providers <= 1:
                order_n = (page - 1) * limit
                for tx in block.transactions:
                    if isinstance(tx, BitcoinTransaction):
                        self.cache.store_transaction(tx, order_n, commit=False)
                    order_n += 1
                self.cache.commit()
            self.complete = True if len(block.transactions) == block.tx_count else False
            self.cache.store_block(block)
        return block

    async def getrawblock(self, blockid: str | int):
        """
        Get raw block as hexadecimal string for block with specified hash or block height.

        Not many providers offer this option, and it can be slow, so it is advised to use a local client such
        as bitcoind.

        :param blockid: Block hash or block height
        :type blockid: str, int

        :return str:
        """
        return await self._provider_execute("getrawblock", blockid)

    async def mempool(self, txid: str = "") -> list:
        """
        Get list of all transaction IDs in the current mempool

        A full list of transactions ID's will only be returned if a bcoin or bitcoind client is available. Otherwise
        specify the txid option to verify if a transaction is added to the mempool.

        :param txid: Check if transaction with this hash exists in memory pool
        :type txid: str

        :return list:
        """
        return await self._provider_execute("mempool", txid)

    async def getcacheaddressinfo(self, address: str) -> dict[str, str]:
        """
        Get address information from cache. I.e. balance, number of transactions, number of utox's, etc

        Cache will only be filled after all transactions for a specific address are retrieved (with get_transactions ie)

        :param address: address string
        :type address: str

        :return dict:
        """
        addr_dict = {"address": address}
        addr_rec = await self.cache.getaddress(address)

        if isinstance(addr_rec, DbCacheAddress):
            addr_dict["balance"] = addr_rec.balance
            addr_dict["last_block"] = addr_rec.last_block
            addr_dict["n_txs"] = addr_rec.n_txs
            addr_dict["n_utxos"] = addr_rec.n_utxos

        return addr_dict

    async def isspent(self, txid: str, output_n: int) -> bool:
        """
        Check if the output with provided transaction ID and output number is spent.

        :param txid: Transaction ID hex
        :type txid: str
        :param output_n: Output number
        :type output_n: int

        :return bool:
        """
        t = await self.cache.get_transaction(bytes.fromhex(txid))
        if t and len(t.outputs) > output_n and t.outputs[output_n].spent is not None:
            return t.outputs[output_n].spent
        else:
            return bool(await self._provider_execute("isspent", txid, output_n))

    async def getinfo(self):
        """
        Returns info about current network. Such as difficulty, latest block, mempool size and network hashrate.

        :return dict:
        """
        return await self._provider_execute("getinfo")

    async def getinputvalues(self, t: GenericTransaction) -> GenericTransaction:
        """
        Retrieve values for transaction inputs for given Transaction.

        Raw transactions as stored on the blockchain do not contain the input values but only the previous
        transaction hash and index number. This method retrieves the previous transaction and reads the value.

        :param t: Transaction
        :type t: Transaction

        :return Transaction:
        """
        prev_txs = []
        for i in t.inputs:
            if not i.value:
                if i.prev_txid not in prev_txs:
                    prev_t = await self.get_transactions([i.prev_txid.hex()])
                else:
                    prev_t = [t for t in prev_txs if t.txid == i.prev_txid][0]
                i.value = prev_t.outputs[i.output_n_int].value
        return t


class Cache:
    """
    Store transaction, utxo and address information in database to increase speed and avoid duplicate calls to
    service providers.

    Once confirmed a transaction is immutable so we have to fetch it from a service provider only once. When checking
    for new transactions or utxo's for a certain address we only have to check the new blocks.

    This class is used by the Service class and normally you won't need to access it directly.

    """

    def __init__(self, network: Network, db_uri: str = ""):
        """
        Open Cache class

        :param network: Specify network used
        :type network: Network
        :param db_uri: Database to use for caching
        :type db_uri: str
        """

        # self.session = None
        # if SERVICE_CACHING_ENABLED:
        #     self.cache = DbCache(db_uri)

        self.db_uri = db_uri
        self.network = network
        self.cache: DbCache | None = None
        self.cache_started = False

    def __exit__(self):
        try:
            self.session.close()
        except Exception:
            pass

    async def cache_enabled(self) -> bool:
        """
        Check if caching is enabled. Returns False if SERVICE_CACHING_ENABLED is False or no session is defined.

        :return bool:
        """
        if not self.cache_started and SERVICE_CACHING_ENABLED:
            self.cache = await DbCache.start(self.db_uri)
            self.cache_started = True

        if not SERVICE_CACHING_ENABLED or not self.cache.connection_possible:
            return False

        return True

    # def commit(self):
    #     """
    #     Commit queries in self.session. Rollback if commit fails.

    #     :return:
    #     """
    #     if not self.session:
    #         return
    #     try:
    #         self.session.commit()
    #     except Exception:
    #         self.session.rollback()
    #         raise

    @staticmethod
    def _parse_db_transaction(db_tx: DbCacheTransaction):
        # probably store this on provider or something
        if db_tx.network_name == "flux":
            klass = FluxTransaction
        else:
            klass = BitcoinTransaction

        t = klass(
            locktime=db_tx.locktime,
            version=db_tx.version,
            network=db_tx.network_name,
            fee=db_tx.fee,
            txid=db_tx.txid.hex(),
            date=db_tx.date,
            confirmations=db_tx.confirmations,
            block_height=db_tx.block_height,
            status="confirmed",
            witness_type=db_tx.witness_type.value,
        )

        for n in db_tx.nodes:
            if n.is_input:
                if n.ref_txid == b"\00" * 32:
                    t.coinbase = True
                t.add_input(
                    n.ref_txid.hex(),
                    n.ref_index_n,
                    unlocking_script=n.script,
                    address=n.address,
                    sequence=n.sequence,
                    value=n.value,
                    index_n=n.index_n,
                    witnesses=n.witnesses,
                    strict=False,
                )
            else:
                t.add_output(
                    n.value,
                    n.address,
                    lock_script=n.script,
                    spent=n.spent,
                    output_n=n.index_n,
                    spending_txid=None if not n.ref_txid else n.ref_txid.hex(),
                    spending_index_n=n.ref_index_n,
                    strict=False,
                )

        t.update_totals()
        t.size = len(t.raw())
        t.calc_weight_units()
        _logger.info("Retrieved transaction %s from cache" % t.txid)
        return t

    @test_cache
    async def get_transaction(self, txid: bytes) -> GenericTransaction:
        """
        Get transaction from cache. Returns False if not available

        :param txid: Transaction identification hash
        :type txid: bytes

        :return Transaction: A single Transaction object
        """
        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheTransaction).filter_by(
                    txid=txid, network_name=self.network.name
                )
            )

            db_tx = res.first()

            if not db_tx:
                return False

            await db_tx.awaitable_attrs.nodes

        # ??? removed this
        # db_tx.txid = txid

        t = self._parse_db_transaction(db_tx)

        if t.block_height:
            t.confirmations = (await self.blockcount() - t.block_height) + 1

        return t

    @test_cache
    async def getaddress(self, address: str) -> DbCacheAddress | None:
        """
        Get address information from cache, with links to transactions and utxo's and latest update information.

        :param address: Address string
        :type address: str

        :return DbCacheAddress: An address cache database object
        """
        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheAddress).filter_by(
                    address=address, network_name=self.network.name
                )
            )

        return res.first()

    @test_cache
    async def get_transactions(
        self, address: str, after_txid: bytes = b"", limit: int = MAX_TRANSACTIONS
    ) -> list[GenericTransaction]:
        """
        Get transactions from cache. Returns empty list if no transactions are found or caching is disabled.

        :param address: Address string
        :type address: str
        :param after_txid: Transaction ID of last known transaction. Only check for transactions after given tx id. Default: Leave empty to return all transaction. If used only provide a single address
        :type after_txid: bytes
        :param limit: Maximum number of transactions to return
        :type limit: int

        :return list: List of Transaction objects
        """
        db_addr = await self.getaddress(address)

        if not db_addr:
            return []

        async with self.cache.get_session() as session:
            if after_txid:
                res = await session.scalars(
                    select(DbCacheTransaction).filter_by(
                        txid=after_txid, network_name=self.network.name
                    )
                )
                target_tx = res.first()

                if target_tx and db_addr.last_block and target_tx.block_height:
                    res = await session.scalars(
                        select(DbCacheTransaction)
                        .join(DbCacheTransactionNode)
                        .filter(
                            DbCacheTransactionNode.address == address,
                            DbCacheTransaction.block_height >= target_tx.block_height,
                            DbCacheTransaction.block_height <= db_addr.last_block,
                        )
                        .order_by(
                            DbCacheTransaction.block_height,
                            DbCacheTransaction.order_n,
                        )
                    )
                    db_txs = res.all()

                    db_txs2 = []
                    # this seems fucked, add breakpoint
                    for d in db_txs:
                        db_txs2.append(d)
                        if d.txid == after_txid:
                            db_txs2 = []
                    db_txs = db_txs2
                else:
                    return []
            else:
                res = await session.scalars(
                    select(DbCacheTransaction)
                    .join(DbCacheTransactionNode)
                    .filter(DbCacheTransactionNode.address == address)
                    .order_by(
                        DbCacheTransaction.block_height, DbCacheTransaction.order_n
                    )
                )
                db_txs = res.all()

            for db_tx in db_txs:
                await db_tx.awaitable_attrs.nodes

        txs: list[GenericTransaction] = []
        for db_tx in db_txs:
            t = self._parse_db_transaction(db_tx)

            if t:
                if t.block_height:
                    t.confirmations = (await self.blockcount() - t.block_height) + 1
                txs.append(t)
                if len(txs) >= limit:
                    break

        return txs

    @test_cache
    async def getblocktransactions(
        self, height: int, page: int, limit: int
    ) -> list[GenericTransaction]:
        """
        Get range of transactions from a block

        :param height: Block height
        :type height: int
        :param page: Transaction page
        :type page: int
        :param limit: Number of transactions per page
        :type limit: int

        :return:
        """
        n_from = (page - 1) * limit
        n_to = page * limit

        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheTransaction).filter(
                    DbCacheTransaction.block_height == height,
                    DbCacheTransaction.order_n >= n_from,
                    DbCacheTransaction.order_n < n_to,
                )
            )
            db_txs = res.all()

        txs: list[GenericTransaction] = []
        for db_tx in db_txs:
            t = self._parse_db_transaction(db_tx)
            # why if here?
            if t:
                txs.append(t)
        return txs

    @test_cache
    async def getrawtransaction(self, txid: bytes) -> str | None:
        """
        Get a raw transaction string from the database cache if available

        :param txid: Transaction identification hash
        :type txid: bytes

        :return str: Raw transaction as hexstring
        """
        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheTransaction).filter_by(
                    txid=txid, network_name=self.network.name
                )
            )
            tx = res.first()

        if not tx:
            return

        t = self._parse_db_transaction(tx)
        return t.raw_hex()

    @test_cache
    async def getutxos(
        self, address: str, after_txid: bytes | None = None
    ) -> list[dict]:
        """
        Get list of unspent outputs (UTXO's) for specified address from database cache.

        Sorted from old to new, so highest number of confirmations first.

        :param address: Address string
        :type address: str
        :param after_txid: Transaction ID of last known transaction. Only check for utxos after given tx id. Default: Leave empty to return all utxos.
        :type after_txid: bytes

        :return dict: UTXO's per address
        """
        async with self.cache.get_session() as session:
            db_utxos = await session.execute(
                select(
                    DbCacheTransactionNode.spent,
                    DbCacheTransactionNode.index_n,
                    DbCacheTransactionNode.value,
                    DbCacheTransaction.confirmations,
                    DbCacheTransaction.block_height,
                    DbCacheTransaction.fee,
                    DbCacheTransaction.date,
                    DbCacheTransaction.txid,
                )
                .join(DbCacheTransaction.nodes)
                .order_by(DbCacheTransaction.block_height, DbCacheTransaction.order_n)
                .filter(
                    DbCacheTransactionNode.address == address,
                    DbCacheTransactionNode.is_input == False,
                    DbCacheTransaction.network_name == self.network.name,
                )
            )

        utxos: list[dict] = []
        for db_utxo in db_utxos:
            if db_utxo.spent is False:
                utxos.append(
                    {
                        "address": address,
                        "txid": db_utxo.txid.hex(),
                        "confirmations": db_utxo.confirmations,
                        "output_n": db_utxo.index_n,
                        "input_n": 0,
                        "block_height": db_utxo.block_height,
                        "fee": db_utxo.fee,
                        "size": 0,
                        "value": db_utxo.value,
                        "script": "",
                        "date": db_utxo.date,
                    }
                )
            elif db_utxo.spent is None:
                return utxos

            if db_utxo.txid == after_txid:
                # this makes no sense
                utxos = []

        return utxos

    @test_cache
    async def estimatefee(self, blocks) -> int | None:
        """
        Get fee estimation from cache for confirmation within specified amount of blocks.

        Stored in cache in three groups: low, medium and high fees.

        :param blocks: Expected confirmation time in blocks.
        :type blocks: int

        :return int: Fee in the smallest network denominator (satoshi)
        """
        if blocks <= 1:
            varname = "fee_high"
        elif blocks <= 5:
            varname = "fee_medium"
        else:
            varname = "fee_low"

        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheVars)
                .filter_by(varname=varname, network_name=self.network.name)
                .filter(DbCacheVars.expires > datetime.now())
            )

        if dbvar := res.first():
            return int(dbvar.value)

    @test_cache
    async def blockcount(self, never_expires=False) -> int | None:
        """
        Get number of blocks on the current network from cache if recent data is available.

        :param never_expires: Always return latest blockcount found. Can be used to avoid return to old blocks if service providers are not up-to-date.
        :type never_expires: bool

        :return int:
        """
        async with self.cache.get_session() as session:
            stmt = select(DbCacheVars).filter_by(
                varname="blockcount", network_name=self.network.name
            )
            if not never_expires:
                stmt = stmt.filter(DbCacheVars.expires > datetime.now())

            res = await session.scalars(stmt)

        if dbvar := res.first():
            return int(dbvar.value)
        else:
            return 0

    @test_cache
    async def getblock(self, blockid: int | str) -> Block | None:
        """
        Get specific block from database cache.

        :param blockid: Block height or block hash
        :type blockid: int, str

        :return Block:
        """
        async with self.cache.get_session() as session:
            stmt = select(DbCacheBlock)

            if isinstance(blockid, int):
                stmt = stmt.where(height=blockid, network_name=self.network.name)
            else:
                stmt = stmt.where(block_hash=to_bytes(blockid))

            res = await session.scalars(stmt)
            block = res.first()

        if not block:
            return

        b = Block(
            block_hash=block.block_hash,
            height=block.height,
            network=block.network_name,
            merkle_root=block.merkle_root,
            time=block.time,
            nonce=block.nonce,
            version=block.version,
            prev_block=block.prev_block,
            bits=block.bits,
        )
        b.tx_count = block.tx_count
        _logger.info("Retrieved block with height %d from cache" % b.height)
        return b

    @test_cache
    async def store_blockcount(self, blockcount: int | str):
        """
        Store network blockcount in cache for 60 seconds

        :param blockcount: Number of latest block
        :type blockcount: int, str

        :return:
        """
        async with self.cache.get_session() as session:
            dbvar = DbCacheVars(
                varname="blockcount",
                network_name=self.network.name,
                value=str(blockcount),
                type="int",
                expires=datetime.now() + timedelta(seconds=60),
            )
            await session.merge(dbvar)
            await session.commit()

    @test_cache
    async def store_transaction(
        self, t: GenericTransaction, order_n: int | None = None
    ):
        """
        Store transaction in cache. Use order number to determine order in a block

        :param t: Transaction
        :type t: Transaction
        :param order_n: Order in block
        :type order_n: int
        :param commit: Commit transaction to database. Default is True. Can be disabled if a larger number of transactions are added to cache, so you can commit outside this method.
        :type commit: bool

        :return:
        """
        # Only store complete and confirmed transaction in cache
        if not t.txid:  # pragma: no cover
            _logger.info("Caching failure tx: Missing transaction hash")
            return
        elif not t.date or not t.block_height or not t.network:
            _logger.info(
                "Caching failure tx: Incomplete transaction missing date, block height or network info"
            )
            return
        elif not t.coinbase and [i for i in t.inputs if not i.value]:
            _logger.info("Caching failure tx: One the transaction inputs has value 0")
            return
        # TODO: Check if inputs / outputs are complete? script, value, prev_txid, sequence, output/input_n

        txid = bytes.fromhex(t.txid)

        async with self.cache.get_session() as session:
            res = await session.scalars(
                select(DbCacheTransaction.txid).where(DbCacheTransaction.txid == txid)
            )
            found = res.first()

            # breakpoint()
            if found:
                return

            new_tx = DbCacheTransaction(
                txid=txid,
                date=t.date,
                confirmations=t.confirmations,
                block_height=t.block_height,
                network_name=t.network.name,
                fee=t.fee,
                order_n=order_n,
                version=t.version_int,
                locktime=t.locktime,
                witness_type=t.witness_type,
                expiry_height=t.expiry_height,
            )
            session.add(new_tx)

            for i in t.inputs:
                if (
                    i.value is None or i.address is None or i.output_n is None
                ):  # pragma: no cover
                    _logger.info(
                        "Caching failure tx: Input value, address or output_n missing"
                    )
                    return

                witnesses = int_to_varbyteint(len(i.witnesses)) + b"".join(
                    [bytes(varstr(w)) for w in i.witnesses]
                )

                new_node = DbCacheTransactionNode(
                    txid=txid,
                    address=i.address,
                    index_n=i.index_n,
                    value=i.value,
                    is_input=True,
                    ref_txid=i.prev_txid,
                    ref_index_n=i.output_n_int,
                    script=i.unlocking_script,
                    sequence=i.sequence,
                    witnesses=witnesses,
                )
                session.add(new_node)

            for o in t.outputs:
                if (
                    o.value is None or o.address is None or o.output_n is None
                ):  # pragma: no cover
                    _logger.info(
                        "Caching failure tx: Output value, address or output_n missing"
                    )
                    return

                new_node = DbCacheTransactionNode(
                    txid=txid,
                    address=o.address,
                    index_n=o.output_n,
                    value=o.value,
                    is_input=False,
                    spent=o.spent,
                    ref_txid=None
                    if not o.spending_txid
                    else bytes.fromhex(o.spending_txid),
                    ref_index_n=o.spending_index_n,
                    script=o.lock_script,
                )
                session.add(new_node)

            try:
                await session.commit()
                _logger.info(f"Added transaction {t.txid} to cache")

            except IntegrityError:
                _logger.info(
                    f"Rolling back this transaction, already stored for tx: {txid}"
                )
                await session.rollback()

            except Exception as e:  # pragma: no cover
                _logger.warning(f"Caching failure tx: {e}")

    @test_cache
    async def store_utxo(self, txid: str, index_n: int):
        """
        Store utxo in cache. Updates only known transaction outputs for transactions which are fully cached

        :param txid: Transaction ID
        :type txid: str
        :param index_n: Index number of output
        :type index_n: int
        :param commit: Commit transaction to database. Default is True. Can be disabled if a larger number of transactions are added to cache, so you can commit outside this method.
        :type commit: bool

        :return:
        """
        txid = bytes.fromhex(txid)

        async with self.cache.get_session() as session:
            result = await session.execute(
                update(DbCacheTransactionNode)
                .filter(
                    DbCacheTransactionNode.txid == txid,
                    DbCacheTransactionNode.index_n == index_n,
                    DbCacheTransactionNode.is_input == False,
                )
                .where({DbCacheTransactionNode.spent: False})
            )

            try:
                await session.commit()
            except Exception as e:  # pragma: no cover
                _logger.warning(
                    "Caching failure utxo %s:%d: %s" % (txid.hex(), index_n, e)
                )

    @test_cache
    async def store_address(
        self,
        address: str,
        last_block: int | None = None,
        balance: int = 0,
        n_utxos: int | None = None,
        txs_complete: bool = False,
        last_txid: bytes | None = None,
        last_tx_index: int | None = None,
    ):
        """
               Store address information in cache

               :param address: Address string
               :type address: str
               :param last_block: Number or last block retrieved from service provider. For instance if address contains a large number of transactions and they will be retrieved in more then one request.
               :type last_block: int
               :param balance: Total balance of address in sathosis, or smallest network detominator
               :type balance: int
               :param n_utxos: Total number of UTXO's for this address
               :type n_utxos: int
               :param txs_complete: True if all transactions for this address are added to cache
               :type txs_complete: bool
               :param last_txid: Transaction ID of last transaction downloaded from blockchain
               :type last_txid: bytes

        .       :return:
        """
        n_txs = None
        async with self.cache.get_session() as session:
            if txs_complete:
                res = await session.scalars(
                    select(DbCacheTransaction)
                    .join(DbCacheTransactionNode)
                    .filter(DbCacheTransactionNode.address == address)
                )
                n_txs = len(res.all())

                if n_utxos is None:
                    n_utxos = await session.scalars(
                        select(func.count(DbCacheTransactionNode.txid)).where(
                            DbCacheTransactionNode.address == address,
                            DbCacheTransactionNode.spent.is_(False),
                            DbCacheTransactionNode.is_input.is_(False),
                        )
                    )
                    if await session.scalars(
                        select(func.count(DbCacheTransactionNode.txid)).where(
                            DbCacheTransactionNode.address == address,
                            DbCacheTransactionNode.spent.is_(None),
                            DbCacheTransactionNode.is_input.is_(False),
                        )
                    ):
                        n_utxos = None

                if not balance:
                    res = await session.execute(
                        select(
                            DbCacheTransactionNode.is_input,
                            func.sum(DbCacheTransactionNode.value),
                        )
                        .filter(DbCacheTransactionNode.address == address)
                        .group_by(DbCacheTransactionNode.is_input)
                    )
                    plusmin = res.all()

                    # this could really use some explaining
                    balance = (
                        0
                        if not plusmin
                        else sum([(-p[1] if p[0] else p[1]) for p in plusmin])
                    )

            db_addr = await self.getaddress(address)

            # this seems ridiculous
            new_address = DbCacheAddress(
                address=address,
                network_name=self.network.name,
                last_block=last_block
                if last_block
                else getattr(db_addr, "last_block", None),
                balance=balance
                if balance is not None
                else getattr(db_addr, "balance", None),
                n_utxos=n_utxos
                if n_utxos is not None
                else getattr(db_addr, "n_utxos", None),
                n_txs=n_txs if n_txs is not None else getattr(db_addr, "n_txs", None),
                last_txid=last_txid
                if last_txid is not None
                else getattr(db_addr, "last_txid", None),
                last_tx_index=last_tx_index
                if last_tx_index is not None
                else getattr(db_addr, "last_tx_index", None),
            )

            await session.merge(new_address)
            try:
                await session.commit()
            except Exception as e:  # pragma: no cover
                _logger.warning("Caching failure addr: %s" % e)

    @test_cache
    async def store_estimated_fee(self, blocks: int, fee: int):
        """
        Store estimated fee retrieved from service providers in cache.

        :param blocks: Confirmation within x blocks
        :type blocks: int
        :param fee: Estimated fee in Sathosis
        :type fee: int

        :return:
        """
        if blocks <= 1:
            varname = "fee_high"
        elif blocks <= 5:
            varname = "fee_medium"
        else:
            varname = "fee_low"
        dbvar = DbCacheVars(
            varname=varname,
            network_name=self.network.name,
            value=str(fee),
            type="int",
            expires=datetime.now() + timedelta(seconds=600),
        )

        async with self.cache.get_session() as session:
            await session.merge(dbvar)
            await session.commit()

    @test_cache
    async def store_block(self, block: Block):
        """
        Store block in cache database

        :param block: Block
        :type block: Block

        :return:
        """

        # this genesis block thing seems fucked
        if (
            not (
                block.height
                and block.block_hash
                and block.prev_block
                and block.merkle_root
                and block.bits
                and block.version
            )
            and not block.block_hash
            == b"\x00\x00\x00\x00\x00\x19\xd6h\x9c\x08Z\xe1e\x83\x1e\x93O\xf7c\xaeF"
            b"\xa2\xa6\xc1r\xb3\xf1\xb6\n\x8c\xe2o"
        ):  # Bitcoin genesis block
            _logger.info("Caching failure block: incomplete data")
            return

        new_block = DbCacheBlock(
            block_hash=block.block_hash,
            height=block.height,
            network_name=self.network.name,
            version=block.version_int,
            prev_block=block.prev_block,
            bits=block.bits_int,
            merkle_root=block.merkle_root,
            nonce=block.nonce_int,
            time=block.time,
            tx_count=block.tx_count,
        )
        async with self.cache.get_session() as session:
            await session.merge(new_block)

            try:
                await session.commit()
            except Exception as e:  # pragma: no cover
                _logger.warning("Caching failure block: %s" % e)



import logging

from sqlalchemy import delete, select, update

from fluxwallet.db_new import (
    Db,
    DbKey,
    DbKeyMultisigChildren,
    DbTransaction,
    DbTransactionInput,
    DbTransactionOutput,
    DbWallet,
)
from fluxwallet.wallet import Wallet

_logger = logging.getLogger(__name__)


async def wallets_list(db_uri=None, include_cosigners=False):
    """
    List Wallets from database

    :param db_uri: URI of the database
    :type db_uri: str

    :param include_cosigners: Child wallets for multisig wallets are for internal use only and are skipped by default
    :type include_cosigners: bool

    :return dict: Dictionary of wallets defined in database
    """
    async with Db(db_uri=db_uri) as session:
        stmt_wallets = select(DbWallet).order_by(DbWallet.id)

        db_wallets = await session.scalars(stmt_wallets)

        wallets = []
        for w in db_wallets:
            if w.parent_id and not include_cosigners:
                continue
            wallets.append(
                {
                    "id": w.id,
                    "name": w.name,
                    "owner": w.owner,
                    "network": w.network_name,
                    "purpose": w.purpose,
                    "scheme": w.scheme,
                    "main_key_id": w.main_key_id,
                    "parent_id": w.parent_id,
                }
            )

    return wallets


# test this out
# from sqlalchemy import inspect

# def object_as_dict(obj):
#     return {
#         c.key: getattr(obj, c.key)
#         for c in inspect(obj).mapper.column_attrs
#     }

# user = session.query(User).first()

# d = object_as_dict(user)


async def wallet_exists(wallet, db_uri=None):
    """
    Check if Wallets is defined in database

    :param wallet: Wallet ID as integer or Wallet Name as string
    :type wallet: int, str
    :param db_uri: URI of the database
    :type db_uri: str

    :return bool: True if wallet exists otherwise False
    """

    if wallet in [x["name"] for x in await wallets_list(db_uri)]:
        return True
    if isinstance(wallet, int) and wallet in [
        x["id"] for x in await wallets_list(db_uri)
    ]:
        return True
    return False


async def wallet_create_or_open(
    name,
    keys="",
    owner="",
    network=None,
    account_id=0,
    purpose=None,
    scheme="bip32",
    sort_keys=True,
    password="",
    witness_type=None,
    encoding=None,
    multisig=None,
    sigs_required=None,
    cosigner_id=None,
    key_path=None,
    db_uri=None,
):
    """
    Create a wallet with specified options if it doesn't exist, otherwise just open

    Returns Wallet object

    See Wallets class create method for option documentation
    """

    if await wallet_exists(name, db_uri=db_uri):
        if keys or owner or password or witness_type or key_path:
            _logger.warning("Opening existing wallet, extra options are ignored")
        return await Wallet(name, db_uri=db_uri).open()
    else:
        return await Wallet.create(
            name,
            keys,
            owner,
            network,
            account_id,
            purpose,
            scheme,
            sort_keys,
            password,
            witness_type,
            encoding,
            multisig,
            sigs_required,
            cosigner_id,
            key_path,
            db_uri=db_uri,
        )


async def wallet_delete(
    wallet: Wallet | int | str, db_uri: str | None = None, force: bool = False
):
    """
    Delete wallet and associated keys and transactions from the database. If wallet has unspent outputs it raises a
    WalletError exception unless 'force=True' is specified

    :param wallet: Wallet ID as integer or Wallet Name as string
    :type wallet: int, str
    :param db_uri: URI of the database
    :type db_uri: str
    :param force: If set to True wallet will be deleted even if unspent outputs are found. Default is False
    :type force: bool

    :return int: Number of rows deleted, so 1 if successful
    """
    async with Db(db_uri=db_uri) as session:
        if isinstance(wallet, int) or wallet.isdigit():
            res = await session.scalars(select(DbWallet).filter_by(id=wallet))
        else:
            res = await session.scalars(select(DbWallet).filter_by(name=wallet))

        w = res.first()

        if not w:
            raise WalletError("Wallet '%s' not found" % wallet)

        wallet_id = w.id

        # Delete co-signer wallets if this is a multisig wallet
        for cw in await session.scalars(
            select(DbWallet).filter_by(parent_id=wallet_id)
        ):
            await wallet_delete(cw.id, db_uri=db_uri, force=force)

        # Delete keys from this wallet and update transactions (remove key_id)
        ks = await session.scalars(select(DbKey).filter_by(wallet_id=wallet_id))
        if bool([k for k in ks if k.balance and k.is_private]) and not force:
            raise WalletError(
                "Wallet still has unspent outputs. Use 'force=True' to delete this wallet"
            )

        k_ids = [k.id for k in ks]

        statements = []

        statements.append(
            update(DbTransactionOutput)
            .where(DbTransactionOutput.key_id.in_(k_ids))
            .values(key_id=None)
        )

        statements.append(
            update(DbTransactionInput)
            .where(DbTransactionInput.key_id.in_(k_ids))
            .values(key_id=None)
        )

        statements.append(
            delete(DbKeyMultisigChildren).where(
                DbKeyMultisigChildren.parent_id.in_(k_ids)
            )
        )

        statements.append(
            delete(DbKeyMultisigChildren).where(
                DbKeyMultisigChildren.child_id.in_(k_ids)
            )
        )

        statements.append(delete(DbKey).where(DbKey.wallet_id == wallet))

        for statement in statements:
            await session.execute(statement)

        statements = []

        txs = await session.scalars(
            select(DbTransaction).filter_by(wallet_id=wallet_id)
        )

        for tx in txs:
            statements.append(
                delete(DbTransactionOutput).where(
                    DbTransactionOutput.transaction_id == tx.id
                )
            )
            statements.append(
                delete(DbTransactionInput).where(
                    DbTransactionInput.transaction_id == tx.id
                )
            )

        statements.append(
            delete(DbTransaction).where(DbTransaction.wallet_id == wallet_id)
        )

        for statement in statements:
            await session.execute(statement)

        statements = []

        # Unlink transactions from this wallet (remove wallet_id)
        # stmt_unlink_txs = (
        #     update(DbTransaction)
        #     .where(DbTransaction.wallet_id == wallet_id)
        #     .values(wallet_id=None)
        # )

        # await session.execute(stmt_unlink_txs)
        stmt_del_wallet = delete(DbWallet).where(DbWallet.id == wallet_id)

        res = await session.execute(stmt_del_wallet)
        await session.commit()

    _logger.info("Wallet '%s' deleted" % wallet)

    return res


async def wallet_empty(wallet: Wallet | int | str, db_uri: str | None = None):
    """
    Remove all generated keys and transactions from wallet. Does not delete the wallet itself or the masterkey,
    so everything can be recreated.

    :param wallet: Wallet ID as integer or Wallet Name as string
    :type wallet: int, str
    :param db_uri: URI of the database
    :type db_uri: str

    :return bool: True if successful
    """

    async with Db(db_uri=db_uri) as session:
        if isinstance(wallet, int) or wallet.isdigit():
            res = await session.scalars(select(DbWallet).filter_by(id=wallet))
        else:
            res = await session.scalars(select(DbWallet).filter_by(name=wallet))

        w = res.first()

        if not w:
            raise WalletError("Wallet '%s' not found" % wallet)

        wallet_id = w.id

        ks = await session.scalars(
            select(DbKey).where(DbKey.wallet_id == wallet_id, DbKey.parent_id != 0)
        )

        k_ids = [k.id for k in ks]

        statements = []

        statements.append(
            update(DbTransactionOutput)
            .where(DbTransactionOutput.key_id.in_(k_ids))
            .values(key_id=None)
        )

        statements.append(
            update(DbTransactionInput)
            .where(DbTransactionInput.key_id.in_(k_ids))
            .values(key_id=None)
        )

        statements.append(
            delete(DbKeyMultisigChildren).where(
                DbKeyMultisigChildren.parent_id.in_(k_ids)
            )
        )

        statements.append(
            delete(DbKeyMultisigChildren).where(
                DbKeyMultisigChildren.child_id.in_(k_ids)
            )
        )

        for statement in statements:
            await session.execute(statement)

        statements = []

        txs = await session.scalars(
            select(DbTransaction).filter_by(wallet_id=wallet_id)
        )

        for tx in txs:
            statements.append(
                delete(DbTransactionOutput).where(
                    DbTransactionOutput.transaction_id == tx.id
                )
            )
            statements.append(
                delete(DbTransactionInput).where(
                    DbTransactionInput.transaction_id == tx.id
                )
            )

        statements.append(
            delete(DbTransaction).where(DbTransaction.wallet_id == wallet_id)
        )

        for statement in statements:
            await session.execute(statement)

        await session.commit()

        _logger.info("All keys and transactions from wallet '%s' deleted" % wallet)

        return True


async def wallet_delete_if_exists(wallet, db_uri=None, force=False):
    """
    Delete wallet and associated keys from the database. If wallet has unspent outputs it raises a WalletError exception
    unless 'force=True' is specified. If the wallet does not exist return False

    :param wallet: Wallet ID as integer or Wallet Name as string
    :type wallet: int, str
    :param db_uri: URI of the database
    :type db_uri: str
    :param force: If set to True wallet will be deleted even if unspent outputs are found. Default is False
    :type force: bool

    :return int: Number of rows deleted, so 1 if successful
    """

    if await wallet_exists(wallet, db_uri):
        return await wallet_delete(wallet, db_uri, force)
    return False


def normalize_path(path):
    """
    Normalize BIP0044 key path for HD keys. Using single quotes for hardened keys

    >>> normalize_path("m/44h/2p/1'/0/100")
    "m/44'/2'/1'/0/100"

    :param path: BIP0044 key path
    :type path: str

    :return str: Normalized BIP0044 key path with single quotes
    """

    levels = path.split("/")
    npath = ""
    for level in levels:
        if not level:
            raise WalletError("Could not parse path. Index is empty.")
        nlevel = level
        if level[-1] in "'HhPp":
            nlevel = level[:-1] + "'"
        npath += nlevel + "/"
    if npath[-1] == "/":
        npath = npath[:-1]
    return npath



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Base Client
#    © 2016-2022 - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
from urllib.parse import urlencode

import requests

from fluxwallet.keys import Address
from fluxwallet.main import *
from fluxwallet.networks import Network

_logger = logging.getLogger(__name__)


class ClientError(Exception):
    def __init__(self, msg=""):
        self.msg = msg
        _logger.info(msg)

    def __str__(self):
        return self.msg


class BaseClient(object):
    def __init__(
        self,
        network: str,
        provider: str,
        base_url: str,
        denominator: int,
        api_key: str = "",
        provider_coin_id: str = "",
        network_overrides=None,
        timeout=TIMEOUT_REQUESTS,
        latest_block=None,
        strict=True,
    ):
        try:
            self.network = network
            if not isinstance(network, Network):
                self.network = Network(network)
            self.provider = provider
            self.base_url = base_url
            self.resp = None
            self.units = denominator
            self.api_key = api_key
            self.provider_coin_id = provider_coin_id
            self.network_overrides = {}
            self.timeout = timeout
            self.latest_block = latest_block
            if network_overrides is not None:
                self.network_overrides = network_overrides
            self.strict = strict
        except Exception:
            raise ClientError("This Network is not supported by %s Client" % provider)

    def request(
        self,
        url_path: str,
        variables: dict = {},
        method: str = "get",
        secure: bool = True,
        post_data: str = "",
        base: str | None = None,
    ):
        url_vars = ""
        url = f"{self.base_url}{url_path}" if base is None else f"{base}{url_path}"

        if not url or not self.base_url:
            raise ClientError("No (complete) url provided: %s" % url)

        headers = {
            "User-Agent": f"fluxwallet/{FLUXWALLET_VERSION}",
            "Accept": "application/json",
            # 'Content-Type': 'application/json',
            # "Referrer": "https://www.github.com/1200wd/fluxwallet",
        }

        if method == "get":
            if variables:
                url_vars = "?" + urlencode(variables)
            url += url_vars
            log_url = url if "@" not in url else url.split("@")[1]
            _logger.info("Url get request %s" % log_url)
            self.resp = requests.get(
                url, timeout=self.timeout, verify=secure, headers=headers
            )
        elif method == "post":
            log_url = url if "@" not in url else url.split("@")[1]
            _logger.info("Url post request %s" % log_url)

            try:
                self.resp = requests.post(
                    url,
                    json=dict(variables),
                    data=post_data,
                    timeout=self.timeout,
                    verify=secure,
                    headers=headers,
                )
            except Exception as e:
                print("exception")
                print(repr(e))
                exit(0)
        resp_text = self.resp.text
        if len(resp_text) > 1000:
            resp_text = self.resp.text[:970] + "... truncated, length %d" % len(
                resp_text
            )
        _logger.debug("Response [%d] %s" % (self.resp.status_code, resp_text))
        log_url = url if "@" not in url else url.split("@")[1]
        if self.resp.status_code == 429:
            raise ClientError(
                "Maximum number of requests reached for %s with url %s, response [%d] %s"
                % (self.provider, log_url, self.resp.status_code, resp_text)
            )
        elif not (self.resp.status_code == 200 or self.resp.status_code == 201):
            raise ClientError(
                "Error connecting to %s on url %s, response [%d] %s"
                % (self.provider, log_url, self.resp.status_code, resp_text)
            )
        try:
            if not self.resp.apparent_encoding and not self.resp.encoding:
                return self.resp.content
            return json.loads(self.resp.text)
        except ValueError or json.decoder.JSONDecodeError:
            return self.resp.text

    def _address_convert(self, address):
        if not isinstance(address, Address):
            return Address.parse(
                address,
                network_overrides=self.network_overrides,
                network=self.network.name,
            )

    def _addresslist_convert(self, addresslist):
        addresslistconv = []
        for address in addresslist:
            addresslistconv.append(self._address_convert(address))
        return addresslistconv
